#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import List, Union, Callable


class RotationAngle(object):

    def __init__(self, angle_relationship: Callable, pauli_label: List,
                 pauli_angle: Union[int, float] = None):

        self._angle = angle_relationship
        self.pauli_label = pauli_label
        self.pauli_angle = pauli_angle

    @property
    def rotation_angle(self) -> Union[int, float]:

        return self._angle(self.pauli_angle)



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
from time import time
from typing import Optional, List
# IBM Qiskit imports
from qiskit import QuantumCircuit, QuantumRegister, execute
from qiskit.providers.ibmq.job import (IBMQJobApiError, IBMQJobInvalidStateError,
                                       IBMQJobFailureError, IBMQJobTimeoutError)
from qiskit.circuit import Parameter

from ...devices import DeviceQiskit
from ...basebackend import QAOABaseBackendShotBased, QAOABaseBackendCloud, QAOABaseBackendParametric
from ...qaoa_parameters.baseparams import QAOACircuitParams, QAOAVariationalBaseParams
from ...utilities import flip_counts


# add support to perform error mitigation for a future version
# from qiskit.ignis.mitigation.measurement import (complete_meas_cal,
#                                                  CompleteMeasFitter)


class QAOAQiskitQPUBackend(QAOABaseBackendParametric, QAOABaseBackendCloud, QAOABaseBackendShotBased):
    """
    A QAOA simulator as well as for real QPU using qiskit as the backend

    Parameters
    ----------
    device: `DeviceQiskit`
        An object of the class ``DeviceQiskit`` which contains the credentials
        for accessing the QPU via cloud and the name of the device.
    circuit_params: `QAOACircuitParams`
        An object of the class ``QAOACircuitParams`` which contains information on 
        circuit construction and depth of the circuit.
    n_shots: `int`
        The number of shots to be taken for each circuit.
    prepend_state: `QuantumCircuit`
        The state prepended to the circuit.
    append_state: `QuantumCircuit`
        The state appended to the circuit.
    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the 
        QAOA part of the circuit.
    cvar_alpha: `float`
        The value of alpha for the CVaR method.
    """

    def __init__(self,
                 circuit_params: QAOACircuitParams,
                 device: DeviceQiskit,
                 n_shots: int,
                 prepend_state: Optional[QuantumCircuit],
                 append_state: Optional[QuantumCircuit],
                 init_hadamard: bool,
                 qubit_layout: List[int] = [],
                 cvar_alpha: float = 1):

        QAOABaseBackendShotBased.__init__(self,
                                          circuit_params,
                                          n_shots,
                                          prepend_state,
                                          append_state,
                                          init_hadamard,
                                          cvar_alpha)
        QAOABaseBackendCloud.__init__(self, device)

        self.qureg = QuantumRegister(self.n_qubits)
        self.qubit_layout = self.circuit_params.qureg if qubit_layout == [] else qubit_layout

        if self.prepend_state:
            assert self.n_qubits >= len(prepend_state.qubits), "Cannot attach a bigger circuit" \
                                                               "to the QAOA routine"

        if self.device.provider_connected and self.device.qpu_connected:
            self.backend_qpu = self.device.backend_device
        elif self.device.provider_connected and self.device.qpu_connected in [False, None]:
            raise Exception(
                'Connection to IBMQ was made. Error connecting to the specified backend.')
        else:
            raise Exception('Error connecting to IBMQ.')
        # For parametric circuits
        self.parametric_circuit = self.parametric_qaoa_circuit

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> QuantumCircuit:
        """
        The final QAOA circuit to be executed on the QPU.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        qaoa_circuit: `QuantumCircuit`
            The final QAOA circuit after binding angles from variational parameters.
        """

        angles_list = self.obtain_angles_for_pauli_list(
            self.abstract_circuit, params)
        memory_map = dict(zip(self.qiskit_parameter_list, angles_list))
        new_parametric_circuit = self.parametric_circuit.bind_parameters(
            memory_map)
        return new_parametric_circuit

    @property
    def parametric_qaoa_circuit(self) -> QuantumCircuit:
        """
        Creates a parametric QAOA circuit, given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit. To do this, you will need to subsequently execute the command self.eng.flush().

        Parameters
        ----------
            params:
                Object of type QAOAVariationalBaseParams
        """
        # self.reset_circuit()
        parametric_circuit = QuantumCircuit(self.qureg)

        if self.prepend_state:
            parametric_circuit = parametric_circuit.compose(self.prepend_state)
        # Initial state is all |+>
        if self.init_hadamard:
            parametric_circuit.h(self.qureg)

        self.qiskit_parameter_list = []
        for each_gate in self.abstract_circuit:
            angle_param = Parameter(str(each_gate.pauli_label))
            self.qiskit_parameter_list.append(angle_param)
            each_gate.rotation_angle = angle_param
            decomposition = each_gate.decomposition('standard')
            # using the list above, construct the circuit
            for each_tuple in decomposition:
                gate = each_tuple[0]()
                parametric_circuit = gate.apply_ibm_gate(*each_tuple[1],parametric_circuit)

        if self.append_state:
            parametric_circuit = parametric_circuit.compose(self.append_state)
        parametric_circuit.measure_all()

        return parametric_circuit

    def get_counts(self, params: QAOAVariationalBaseParams) -> dict:
        """
        Execute the circuit and obtain the counts

        Parameters
        ----------
        params: QAOAVariationalBaseParams
            The QAOA parameters - an object of one of the parameter classes, containing 
            variable parameters.

        Returns
        -------
            A dictionary with the bitstring as the key and the number of counts 
            as its value.
        """

        circuit = self.qaoa_circuit(params)

        job_state = False
        no_of_job_retries = 0
        max_job_retries = 5

        while job_state == False:
            job = execute(circuit, backend=self.backend_qpu,
                          shots=self.n_shots, initial_layout=self.qubit_layout)

            api_contact = False
            no_of_api_retries = 0
            max_api_retries = 5

            while api_contact == False:
                try:
                    self.job_id = job.job_id()
                    counts = job.result().get_counts()
                    api_contact = True
                    job_state = True
                except (IBMQJobApiError, IBMQJobTimeoutError):
                    print("There was an error when trying to contact the IBMQ API.")
                    job_state = True
                    no_of_api_retries += 1
                    time.sleep(5)
                except (IBMQJobFailureError, IBMQJobInvalidStateError):
                    print(
                        "There was an error with the state of the Job in IBMQ.")
                    no_of_job_retries += 1
                    break

                if no_of_api_retries >= max_api_retries:
                    raise ConnectionError(
                        "Number of API Retries exceeded Maximum allowed.")

            if no_of_job_retries >= max_job_retries:
                raise ConnectionError(
                    "An Error Occurred with the Job(s) sent to IBMQ.")

        # Expose counts
        counts_flipped = flip_counts(counts)
        self.measurement_outcomes = counts_flipped
        return counts_flipped

    def circuit_to_qasm(self, params: QAOAVariationalBaseParams) -> str:
        """
        A method to convert the entire QAOA `QuantumCircuit` object into 
        a OpenQASM string
        """
        raise NotImplementedError()
        # qasm_string = self.qaoa_circuit(params).qasm(formatted=True)
        # return qasm_string

    def reset_circuit(self):
        """
        Reset self.circuit after performing a computation
        """
        raise NotImplementedError()



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import List, Dict, Optional, Union
import numpy as np
import qiskit
from scipy.optimize._minimize import MINIMIZE_METHODS
import scipy
import numpy

from openqaoa.basebackend import QuantumCircuitBase
from openqaoa.devices import SUPPORTED_LOCAL_SIMULATORS
from .parameters import Parameters


ALLOWED_PARAM_TYPES = ['standard', 'standard_w_bias', 'extended', 'fourier',
                       'fourier_extended', 'fourier_w_bias', 'annealing']
ALLOWED_INIT_TYPES = ['rand', 'ramp', 'custom']
ALLOWED_MIXERS = ['x', 'xy']
ALLOWED_MINIMIZATION_METHODS = MINIMIZE_METHODS

ALLOWED_QVM_DEVICES = ['Aspen-11', 'Aspen-M-1']
ALLOWED_QVM_DEVICES.extend(f'{n}q-qvm' for n in range(2, 80))

ALLOWED_LOCAL_SIMUALTORS = SUPPORTED_LOCAL_SIMULATORS
ALLOWED_IMBQ_GLOBAL =  ['ibmq_qasm_simulator', 'ibmq_armonk', 'ibmq_santiago', 'ibmq_bogota', 'ibmq_lima', 'ibmq_belem', 'ibmq_quito', 'simulator_statevector', 'simulator_mps', 'simulator_extended_stabilizer', 'simulator_stabilizer', 'ibmq_manila']
ALLOWED_DEVICES  = ALLOWED_LOCAL_SIMUALTORS + ALLOWED_QVM_DEVICES + ALLOWED_IMBQ_GLOBAL


class CircuitProperties(Parameters):
    """
    Tunable properties of the QAOA circuit to be specified by the user
    """

    def __init__(self,
                 param_type: str = 'standard',
                 init_type: str = 'ramp',
                 qubit_register: List = [],
                 p: int = 1,
                 q: Optional[int] = 1,
                 annealing_time: Optional[float] = None,
                 linear_ramp_time: Optional[float] = None,
                 variational_params_dict: Optional[dict] = {},
                 mixer_hamiltonian: Optional[str] = 'x',
                 mixer_qubit_connectivity: Optional[Union[List[list],
                                                          List[tuple], str]] = None,
                 mixer_coeffs: Optional[float] = None,
                 seed: Optional[int] = None):

        self.param_type = param_type
        self.init_type = init_type
        self.qubit_register = qubit_register
        self.p = p
        self.q = q if param_type.lower() in ['fourier','fourier_extended', 'fourier_w_bias'] else None
        self.variational_params_dict = variational_params_dict
        self.annealing_time = annealing_time if annealing_time is not None else 0.7*self.p
        self.linear_ramp_time = linear_ramp_time if linear_ramp_time is not None else 0.7*self.p
        self.mixer_hamiltonian = mixer_hamiltonian
        if self.mixer_hamiltonian.lower() == 'xy':
            self.mixer_qubit_connectivity = mixer_qubit_connectivity if mixer_qubit_connectivity is not None else 'full'
        else:
            self.mixer_qubit_connectivity = None
        self.mixer_coeffs = mixer_coeffs
        self.seed = seed

    @property
    def param_type(self):
        return self._param_type

    @param_type.setter
    def param_type(self, value):
        if value not in ALLOWED_PARAM_TYPES:
            raise ValueError(
                f"param_type {value} is not recognised. Please use {ALLOWED_PARAM_TYPES}")
        self._param_type = value

    @property
    def init_type(self):
        return self._init_type

    @init_type.setter
    def init_type(self, value):
        if value not in ALLOWED_INIT_TYPES:
            raise ValueError(
                f"init_type {value} is not recognised. Please use {ALLOWED_INIT_TYPES}")
        self._init_type = value

    @property
    def mixer_hamiltonian(self):
        return self._mixer_hamiltonian

    @mixer_hamiltonian.setter
    def mixer_hamiltonian(self, value):
        if value not in ALLOWED_MIXERS:
            raise ValueError(
                f"mixer_hamiltonian {value} is not recognised. Please use {ALLOWED_MIXERS}")
        self._mixer_hamiltonian = value

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, value):
        if value <= 0:
            raise ValueError(
                f"Number of layers `p` cannot be smaller or equal to zero. Value {value} was provided")
        self._p = value

    @property
    def annealing_time(self):
        return self._annealing_time

    @annealing_time.setter
    def annealing_time(self, value):
        if value <= 0:
            raise ValueError(
                f"The annealing time `annealing_time` cannot be smaller or equal to zero. Value {value} was provided")
        self._annealing_time = value


    # @property
    # def mixer_qubit_connectivity(self):
    #     return self._mixer_qubit_connectivity

    # @annealing_time.setter
    # def mixer_qubit_connectivity(self, value):
    #     print(value)
    #     if (self.mixer_hamiltonian != 'xy') and (value != None):
    #         self._mixer_qubit_connectivity = None
    #         raise ValueError(f"mixer_qubit_connectivity can be used if and only if `mixer_hamiltonian` is set to `xy`")
    #     else:
    #         print(value)
    #         self._mixer_qubit_connectivity = value

class BackendProperties(Parameters):
    """
    Choose the backend on which to run the QAOA circuits

    Parameters
    ----------
    device: DeviceBase
        The device to use for the backend.
    prepend_state: Union[openqaoa.basebackend.QuantumCircuitBase,numpy.ndarray(complex)]
        The state prepended to the circuit.
    append_state: Union[QuantumCircuitBase,numpy.ndarray(complex)]
        The state appended to the circuit.
    init_hadamard: bool
        Whether to apply a Hadamard gate to the beginning of the 
        QAOA part of the circuit.
    n_shots: int
        The number of shots to be used for the shot-based computation.
    cvar_alpha: float
        The value of the CVaR parameter.
    noise_model: NoiseModel
        The `qiskit` noise model to be used for the shot-based simulator.
    qubit_layout: Union[List[int], numpy.ndarray]
        Mapping from physical to logical qubit indices, used to eventually 
        construct the quantum circuit.  For example, for a system composed by 3 qubits
       `qubit_layout=[1,3,2]`, maps `1<->0`, `3<->1`, `2<->2`, where the left hand side is the physical qubit 
        and the right hand side is the logical qubits
    qiskit_simulation_method: str
        Specify the simulation method to use with the `qiskit.AerSimulator`
    seed_simulator: int
        Specify a seed for `qiskit` simulators
    active_reset: bool
        To use the active_reset functionality on Rigetti backends through QCS
    rewiring: str
        Specify the rewiring strategy for compilation for Rigetti QPUs through QCS
    disable_qubit_rewiring: bool
        enable/disbale qubit rewiring when accessing QPUs via the AWS `braket`
    """

    def __init__(self,
                 prepend_state: Optional[Union[QuantumCircuitBase,
                                               List[complex], np.ndarray]] = None,
                 append_state: Optional[Union[QuantumCircuitBase,
                                              np.ndarray]] = None,
                 init_hadamard: bool = True,
                 n_shots: int = 100,
                 cvar_alpha: float = 1,
                 noise_model = None,
                 qubit_layout: Optional[Union[List[int], np.ndarray]] = None,
                 qiskit_simulation_method: Optional[str] = None,
                 seed_simulator: Optional[int] = None,
                 active_reset: Optional[bool] = None,
                 rewiring: Optional[str] = None,
                 disable_qubit_rewiring: Optional[bool] = None
                 ):
        
        self.init_hadamard = init_hadamard
        self.n_shots = n_shots
        self.prepend_state = prepend_state
        self.append_state = append_state
        self.cvar_alpha = cvar_alpha
        self.noise_model = noise_model
        self.qubit_layout = qubit_layout
        self.seed_simulator = seed_simulator
        self.qiskit_simulation_method = qiskit_simulation_method
        self.active_reset = active_reset
        self.rewiring = rewiring
        self.disable_qubit_rewiring = disable_qubit_rewiring
        
    # @property
    # def cvar_alpha(self):
    #     return self._cvar_alpha

    # @cvar_alpha.setter
    # def cvar_alpha(self, value):
    #     if (value <0) or (value>1) :
    #         raise ValueError(
    #             f"cvar_alpha must be between 0 and 1. Received {value}.")
    #     self._cvar_alpha = value


class ClassicalOptimizer(Parameters):
    """
    The classical optimizer for the QAOA optimization routine 
    of the QAOA circuit parameters.

    Parameters
    ----------
    optimize: bool
        Whether to perform optimization routine on the given QAOA problem
    method: str
        optimization method for QAOA e.g. 'COBYLA'
    maxiter : Optional[int]
        Maximum number of iterations.
    maxfev : Optional[int]
        Maximum number of function evaluations.
    jac: str
        Method to compute the gradient vector. Choose from:
            - ['finite_difference', 'param_shift', 'stoch_param_shift', 'grad_spsa']       
    hess:
        Method to compute the hessian. Choose from:
            - ['finite_difference', 'param_shift', 'stoch_param_shift', 'grad_spsa']
    constraints: `scipy.optimize.LinearConstraints`, `scipy.optimize.NonlinearConstraints` 
        Scipy-based constraints on parameters of optimization 
    bounds: `scipy.scipy.optimize.Bounds`
        Scipy-based bounds on parameters of optimization 
    tol : float
        Tolerance before the optimizer terminates; if `tol` is larger than
        the difference between two steps, terminate optimization.
    optimizer_options: dict
        Dictionary of optimiser-specific arguments, defaults to ``None``.
            stepsize : float
                Step size of each gradient descent step.
            decay : float
                Stepsize decay parameter of RMSProp.
            eps : float
                Small number to prevent division by zero for RMSProp.
            lambd : float
                Small number to regularize QFIM for Natural Gradient Descent.
    jac_options : dict
        Dictionary that specifies gradient-computation options according to method chosen in 'jac'.
    hess_options : dict
        Dictionary that specifies Hessian-computation options according to method chosen in 'hess'.
    optimization_progress : bool
        Returns history of measurement outcomes/wavefunction if `True`. Defaults to `False`.
    cost_progress : bool
        Returns history of cost values if `True`. Defaults to `True`. 
    parameter_log : bool
        Returns history of angles if `True`. Defaults to `True`.
    save_intermediate: bool
        Outputs the jobids and parameters used for each circuit into seperate csv files. Defaults to `False`.
    """

    def __init__(self,
                 optimize: bool = True,
                 method: str = 'cobyla',
                 maxiter: int = 100,
                 maxfev : int = None,
                 jac: str = None,
                 hess: str = None,
                 constraints=None,
                 bounds=None,
                 tol=None,
                 optimizer_options: dict = None,
                 jac_options: dict = None,
                 hess_options: dict = None,
                 optimization_progress: bool = False,
                 cost_progress: bool = True,
                 parameter_log: bool = True, 
                 save_intermediate: bool = False):
        self.optimize = optimize
        self.method = method.lower()
        self.maxiter = maxiter
        self.maxfev = maxfev
        self.jac = jac.lower() if type(jac) == str else jac
        self.hess = hess.lower() if type(hess) == str else hess
        self.constraints = constraints
        self.bounds = bounds
        self.tol = tol
        self.optimizer_options = optimizer_options
        self.jac_options = jac_options
        self.hess_options = hess_options
        self.parameter_log = parameter_log
        self.optimization_progress = optimization_progress
        self.cost_progress = cost_progress
        self.parameter_log = parameter_log
        self.save_intermediate = save_intermediate

    # @property
    # def method(self):
    #     return self._method

    # @method.setter
    # def method(self, value):
    #     if value not in ALLOWED_MINIMIZATION_METHODS:
    #         raise ValueError(
    #             f"method `{value}` is not supported. Please choose between {ALLOWED_MINIMIZATION_METHODS}")
    #     self._method = value




#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from functools import update_wrapper
from logging.config import dictConfig
from re import I
import matplotlib.pyplot as plt
from typing import Type
import numpy as np

from .logger_vqa import Logger
from ..qaoa_parameters.operators import Hamiltonian
from ..utilities import qaoa_probabilities, bitstring_energy


def most_probable_bitstring(cost_hamiltonian, measurement_outcomes):

    mea_out = list(measurement_outcomes.values())
    index_likliest_states = np.argwhere(mea_out == np.max(mea_out))
    # degeneracy = len(index_likliest_states)
    solutions_bitstrings = [
        list(measurement_outcomes.keys())[e[0]] for e in index_likliest_states
    ]

    return {
        "solutions_bitstrings": solutions_bitstrings,
        "bitstring_energy": bitstring_energy(cost_hamiltonian, solutions_bitstrings[0]),
    }


class Result:
    """
    A class to handle the results of QAOA workflows

    Parameters
    ----------
    log: `Logger`
        The raw logger generated from the training vqa part of the QAOA. 
    method: `str`
        Stores the name of the optimisation used by the classical optimiser
    """

    def __init__(
        self, log: Type[Logger], method: Type[str], cost_hamiltonian: Type[Hamiltonian]
    ):

        self.method = method

        self.cost_hamiltonian = cost_hamiltonian

        self.evals = {
            "number of evals": log.func_evals.best[0],
            "jac evals": log.jac_func_evals.best[0],
            "qfim evals": log.qfim_func_evals.best[0],
        }

        self.intermediate = {
            'angles log': np.array(log.param_log.history).tolist(),
            'intermediate cost': log.cost.history,
            'intermediate measurement outcomes':
                log.measurement_outcomes.history,
            'intermediate runs job id': log.job_ids.history
        }

        self.optimized = {
            'optimized angles':
                np.array(log.param_log.best[0]).tolist()
                if log.param_log.best != [] else [],
            'optimized cost':
                log.cost.best[0]
                if log.cost.best != [] else None,
            'optimized measurement outcomes':
                log.measurement_outcomes.best[0]
                if log.measurement_outcomes.best != [] else {},
            'optimized run job id': 
                log.job_ids.best[0] 
                if len(log.job_ids.best) != 0 else []
        }

        self.most_probable_states = most_probable_bitstring(
            cost_hamiltonian, self.get_counts(log.measurement_outcomes.best[0])
        ) if log.measurement_outcomes.best != [] else []

    # def __repr__(self):
    #     """Return an overview over the parameters and hyperparameters
    #     Todo
    #     ----
    #     Split this into ``__repr__`` and ``__str__`` with a more verbose
    #     output in ``__repr__``.
    #     """
    #     string = "Optimization Results:\n"
    #     string += "\tThe solution is " + str(self.solution['degeneracy']) + " degenerate" "\n"
    #     string += "\tThe most probable bitstrings are: " + str(self.solution['bitstring']) + "\n"
    #     string += "\tThe associated cost is: " + str(self.optimized['optimized cost']) + "\n"

    #     return (string)

    @staticmethod
    def get_counts(measurement_outcomes):
        """
        Converts probabilities to counts when the measurement outcomes are a numpy array, that is a state vector

        Parameters
        ----------
        measurement_outcomes: `Union[np.array, dict]`
            The measurement outcome as returned by the Logger. It can either be a statevector or a count dictionary
        Returns
        -------
        `dict`
            The count dictionary obtained either throught the statevector or the actual measurement counts
        """

        if type(measurement_outcomes) == type(np.array([])):
            measurement_outcomes = qaoa_probabilities(measurement_outcomes)

        return measurement_outcomes

    def plot_cost(
        self, figsize=(10, 8), label="Cost", linestyle="--", color="b", ax=None
    ):
        """
        A simpler helper function to plot the cost associated to a QAOA workflow

        Parameters
        ----------
        figsize: `tuple`
            The size of the figure to be plotted. Defaults to (10,8).
        label: `str`
            The label of the cost line, defaults to 'Cost'.
        linestyle: `str`
            The linestyle of the poloit. Defaults to '--'.
        color: `str`
            The color of the line. Defaults to 'b'.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        ax.plot(
            range(
                self.evals["number of evals"]
                - self.evals["jac evals"]
                - self.evals["qfim evals"]
            ),
            self.intermediate["intermediate cost"],
            label=label,
            linestyle=linestyle,
            color=color,
        )

        ax.set_ylabel("Cost")
        ax.set_xlabel("Number of function evaluations")
        ax.legend()
        ax.set_title("Cost history")

        return

    def plot_probabilities(self, n_states_to_keep = None, figsize = (10,8),label='Probability distribution',color='tab:blue', ax=None):

        """
        Helper function to plot the probabilities corresponding to each basis states (with prob != 0) obtained from the optimized result

        Parameters
        ----------
        n_states_to_keep: 'int
            If the user passes a value, the plot will compile with the given value of states. 
            Else,  an upper bound will be calculated depending on the total size of the measurement outcomes.
        figsize: `tuple`
            The size of the figure to be plotted. Defaults to (10,8).
        label: `str`
            The label of the cost line, defaults to 'Probability distribution'.
        color: `str`
            The color of the line. Defaults to 'tab:blue'.
        ax: 'matplotlib.axes._subplots.AxesSubplot'
            Axis on which to plot the graph. Deafults to None
        """

        outcome = self.optimized['optimized measurement outcomes']

        # converting to counts dictionary if outcome is statevector
        if type(outcome) == type(np.array([])):
            outcome = self.get_counts(outcome)
            # setting norm to 1 since it might differ slightly for statevectors due to numerical preicision
            norm = np.float64(1)
        else: 
            # needed to be able to divide the tuple by 'norm'
            norm = np.float64(sum(outcome.values())) 

        # sorting dictionary. adding a callback function to sort by values instead of keys
        # setting reverse = True to be able to obtain the states with highest counts
        outcome_list = sorted(outcome.items(), key=lambda item: item[1], reverse=True)
        states, counts = zip(*outcome_list)

        # normalizing to obtain probabilities
        probs = counts/norm

        # total number of states / number of states with != 0 counts for shot simulators
        total = len(states)

        # number of states that fit without distortion in figure
        upper_bound = 40
        # default fontsize
        font = 'medium'

        if n_states_to_keep:
            if n_states_to_keep > total:
                raise ValueError(f"n_states_to_keep must be smaller or equal than the total number of states in measurement outcome: {total}")
            else:
                if n_states_to_keep>upper_bound:
                    print('number of states_to_keep exceeds the recommended value')
                    font = 'small'

        # if states_to_keep is not given
        else:
            if total > upper_bound:
                n_states_to_keep = upper_bound
            else:
                n_states_to_keep = total
        
        # formatting labels
        labels = [r'$\left|{}\right>$'.format(state) for state in states[:n_states_to_keep]]
        labels.append('rest')

        # represent the bar with the addition of all the remaining probabilites
        rest = sum(probs[n_states_to_keep:])

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        colors = [color for _ in range(n_states_to_keep)] + ['xkcd:magenta']

        ax.bar(labels,np.append(probs[:n_states_to_keep],rest), color=colors)
        ax.set_xlabel('Eigen-State')
        ax.set_ylabel('Probability')
        ax.set_title(label)
        ax.tick_params(axis='x', labelrotation = 75, labelsize=font)
        ax.grid(True, axis='y', linestyle='--')

        print('states kept:', n_states_to_keep)
        return


    def lowest_cost_bitstrings(self, n_bitstrings: int = 1) -> dict:
        """
        Find the minimium energy from cost_hamilonian given a set of measurement
        outcoms

        Parameters
        ----------

        n_bitstrings : int
            Number of the lowest energies bistrings to get

        Returns
        -------
        best_results : dict
            Returns a list of bitstring with the lowest values of the cost Hamiltonian.

        """

        if isinstance(self.optimized["optimized measurement outcomes"], dict):
            measurement_outcomes = self.optimized["optimized measurement outcomes"]
            solution_bitstring = list(measurement_outcomes.keys())
        elif isinstance(self.optimized["optimized measurement outcomes"], np.ndarray):
            measurement_outcomes = self.get_counts(
                self.optimized["optimized measurement outcomes"]
            )
            solution_bitstring = list(measurement_outcomes.keys())
        else:
            raise TypeError(
                f"The measurement outcome {type(self.optimized['optimized measurement outcomes'])} is not valid."
            )
        energies = [
            bitstring_energy(self.cost_hamiltonian, bitstring)
            for bitstring in solution_bitstring
        ]
        args_sorted = np.argsort(energies)
        if n_bitstrings > len(energies):
            n_bitstrings = len(energies)
        
        total_shots = sum(measurement_outcomes.values())
        best_results = {
            "solutions_bitstrings": [
                solution_bitstring[args_sorted[ii]] for ii in range(n_bitstrings)
            ],
            "bitstrings_energies": [
                energies[args_sorted[ii]] for ii in range(n_bitstrings)
            ],
            "probabilities": [
                measurement_outcomes[solution_bitstring[args_sorted[ii]]]/total_shots for ii in range(n_bitstrings)
            ]
        }
        return best_results



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import annotations

import numpy as np
import random

from copy import deepcopy
from .qaoa_parameters.extendedparams import QAOAVariationalExtendedParams 
from .qaoa_parameters.baseparams import QAOAVariationalBaseParams
from .basebackend import QAOABaseBackend
from .optimizers.logger_vqa import Logger


def update_and_compute_expectation(backend_obj: QAOABaseBackend, 
                                   params: QAOAVariationalBaseParams, 
                                   logger: Logger):
    """
    Helper function that returns a callable that takes in a list/nparray of raw parameters.
    This function will handle:
    
        #. Updating logger object with `logger.log_variables`
        #. Updating variational parameters with `update_from_raw` 
        #. Computing expectation with `backend_obj.expectation`
    
    Parameters
    ----------
    backend_obj: QAOABaseBackend
        `QAOABaseBackend` object that contains information about the backend that is being used to perform the QAOA circuit
    params : QAOAVariationalBaseParams
        `QAOAVariationalBaseParams` object containing variational angles. 
    logger: Logger
        Logger Class required to log information from the evaluations required for the jacobian/hessian computation.
    Returns
    -------
    out:
        A callable that accepts a list/array of parameters, and returns the computed expectation value. 
    """
    
    def fun(args):
        current_total_eval = logger.func_evals.best[0]
        current_total_eval += 1
        current_jac_eval = logger.jac_func_evals.best[0]
        current_jac_eval += 1
        logger.log_variables({'func_evals': current_total_eval, 
                              'jac_func_evals': current_jac_eval})
        params.update_from_raw(args)
        return backend_obj.expectation(params)

    return fun

def derivative(backend_obj: QAOABaseBackend, 
               params: QAOAVariationalBaseParams, 
               logger: Logger, 
               derivative_type: str = None, 
               derivative_method: str = None, 
               derivative_options: dict = None):
    """
    Returns a callable function that calculates the gradient according to the specified `gradient_method`.

    Parameters
    ----------
    backend_obj: QAOABaseBackend
        `QAOABaseBackend` object that contains information about the backend that is being used to perform the QAOA circuit
    params : QAOAVariationalBaseParams
        `QAOAVariationalBaseParams` object containing variational angles.
    logger: Logger
        Logger Class required to log information from the evaluations required for the jacobian/hessian computation.
    derivative_type : str
        Type of derivative to compute. Either `gradient` or `hessian`.
    derivative_method : str
        Computational method of the derivative. Either `finite_difference`, `param_shift`, `stoch_param_shift`, or `grad_spsa`.
    derivative_options : dict
        Dictionary containing options specific to each `derivative_method`.
    cost_std : QAOACost
        `QAOACost` object that computes expectation values when executed. Standard parametrisation.
    cost_ext : QAOACost
        `QAOACost` object that computes expectation values when executed. Extended parametrisation. Mainly used to compute parameter shifts at each individual gate, which is summed to recover the parameter shift for a parametrised layer.
    Returns
    -------
    out:
        The callable derivative function of the cost function, generated based on the `derivative_type`, `derivative_method`, and `derivative_options` specified.
    """
    # Default derivative_options used if none are specified.
    default_derivative_options = {"stepsize": 0.00001,
                                  "n_beta_single": -1,
                                  "n_beta_pair": -1,
                                  "n_gamma_single": -1,
                                  "n_gamma_pair": -1}

    derivative_options = {**default_derivative_options, **derivative_options
                          } if derivative_options is not None else default_derivative_options

    # cost_std = derivative_dict['cost_std']
    # cost_ext = derivative_dict['cost_ext']
    params_ext = QAOAVariationalExtendedParams.empty(backend_obj.circuit_params)

    derivative_types = ['gradient', 'hessian']
    assert derivative_type in derivative_types,\
        "Unknown derivative type specified - please choose between " + \
        str(derivative_types)

    derivative_methods = ['finite_difference',
                          'param_shift', 'stoch_param_shift', 'grad_spsa']
    assert derivative_method in derivative_methods,\
        "Unknown derivative computation method specified - please choose between " + \
        str(derivative_methods)
    
    params = deepcopy(params)

    if derivative_type == 'gradient':

        if derivative_method == 'finite_difference':
            out = grad_fd(backend_obj, params, derivative_options, logger)
        elif derivative_method == 'param_shift':
            assert params.__class__.__name__ == 'QAOAVariationalStandardParams', f"{params.__class__.__name__} not supported - only Standard Parametrisation is supported for parameter shift/stochastic parameter shift for now."
            out = grad_ps(backend_obj, params, params_ext, logger)
        elif derivative_method == 'stoch_param_shift':
            assert params.__class__.__name__ == 'QAOAVariationalStandardParams', f"{params.__class__.__name__} not supported - only Standard Parametrisation is supported for parameter shift/stochastic parameter shift for now."
            out = grad_sps(backend_obj, params, params_ext, derivative_options, logger)
        elif derivative_method == 'grad_spsa':
            out = grad_spsa(backend_obj, params, derivative_options, logger)

    elif derivative_type == 'hessian':

        if derivative_method == 'finite_difference':
            out = hessian_fd(backend_obj, params, derivative_options, logger)
        else:
            raise ValueError('Only support hessian derivative method is finite_difference. Your choice: {}'.format(derivative_method))

    return out


def grad_fd(backend_obj, params, gradient_options, logger):
    """
    Returns a callable function that calculates the gradient with the finite difference method.

    PARAMETERS
    ----------
    backend_obj : `QAOABaseBackend`
        backend object that computes expectation values when executed. 

    gradient_options : `dict`
        stepsize : 
            Stepsize of finite difference.

    RETURNS
    -------
    grad_fd_func: `Callable`
        Callable derivative function.
    """

    # Set default value of eta
    eta = gradient_options['stepsize']
    fun = update_and_compute_expectation(backend_obj, params, logger)

    def grad_fd_func(args):

        grad = np.zeros(len(args))

        for i in range(len(args)):
            vect_eta = np.zeros(len(args))
            vect_eta[i] = 1

            # Finite diff. calculation of gradient
            eval_i = fun(args - (eta/2)*vect_eta)
            eval_f = fun(args + (eta/2)*vect_eta)
            grad[i] = (eval_f-eval_i)/eta

        return grad

    return grad_fd_func


def grad_ps(backend_obj, params, params_ext, logger):
    """
    Returns a callable function that calculates the gradient with the parameter shift method.

    PARAMETERS
    ----------
    backend_obj : `QAOABaseBackend`
        backend object that computes expectation values when executed. 

    params : `QAOAVariationalStandardParams`
        variational parameters object, standard parametrisation.

    params_ext : `QAOAVariationalExtendedParams`
        variational parameters object, extended parametrisation.

    RETURNS
    -------
    grad_ps_func:
        Callable derivative function.
    """    
    # TODO : clean up conversion part + handle Fourier parametrisation
    
    fun = update_and_compute_expectation(backend_obj, params_ext, logger)
    
    coeffs_list = params.p*params.mixer_1q_coeffs + params.p*params.mixer_2q_coeffs + \
            params.p*params.cost_1q_coeffs + params.p*params.cost_2q_coeffs

    def grad_ps_func(args):

        # Convert standard to extended parameters before applying parameter shift
        args_ext = params.convert_to_ext(args)
        
        grad_ext = np.zeros(len(args_ext))

        # Apply parameter shifts
        for i in range(len(args_ext)):
            vect_eta = np.zeros(len(args_ext))
            vect_eta[i] = 1
            r = coeffs_list[i]
            grad_ext[i] = r*(fun(args_ext + (np.pi/(4*r))*vect_eta) -
                             fun(args_ext - (np.pi/(4*r))*vect_eta))

        # Convert extended param. gradient form back into std param. form
        
        m1q_entries = grad_ext[:params.p*len(params.mixer_1q_coeffs)]
        m2q_entries = grad_ext[params.p*len(params.mixer_1q_coeffs) : params.p*len(params.mixer_2q_coeffs)]
        c1q_entries = grad_ext[params.p*len(params.mixer_1q_coeffs) + params.p*len(params.mixer_2q_coeffs): params.p*len(params.mixer_1q_coeffs) + params.p*len(params.mixer_2q_coeffs) + params.p*len(params.cost_1q_coeffs)]
        c2q_entries = grad_ext[params.p*len(params.mixer_1q_coeffs) + params.p*len(params.mixer_2q_coeffs) + params.p*len(params.cost_1q_coeffs):]
        subdivided_ext_grad = [m1q_entries, m2q_entries, c1q_entries, c2q_entries]
        
        # Sum up gradients (due to the chain rule), and re-express in standard form.
        mat = np.zeros((4, params.p))
        for i in range(4):  # 4 types of terms
            for j in range(params.p):
                mat[i][j] = np.sum(subdivided_ext_grad[i][j*int(len(subdivided_ext_grad[i]) /
                                   params.p):(j+1)*int(len(subdivided_ext_grad[i])/params.p)])

        grad_std = list(np.sum(mat[:2], axis=0)) + \
            list(np.sum(mat[2:], axis=0))
        
        return np.array(grad_std)

    return grad_ps_func


def grad_sps(backend_obj, params_std, params_ext, gradient_options, logger):
    """
    Returns a callable function that approximates the gradient with the stochastic parameter shift method, which samples (n_beta_single, n_beta_pair, n_gamma_single, n_gamma_pair) gates at each layer instead of all gates. See "Algorithm 4" of https://arxiv.org/pdf/1910.01155.pdf. By convention, (n_beta_single, n_beta_pair, n_gamma_single, n_gamma_pair) = (-1, -1, -1, -1) will sample all gates (which is then equivalent to the full parameter shift rule).

    PARAMETERS
    ----------
    backend_obj : `QAOABaseBackend`
        backend object that computes expectation values when executed. 

    params_std : `QAOAVariationalStandardParams`
        variational parameters object, standard parametrisation.

    params_ext : `QAOAVariationalExtendedParams`
        variational parameters object, extended parametrisation.

    gradient_options :
        n_beta_single : 
            Number of single-qubit mixer gates to sample for the stochastic parameter shift.
        n_beta_pair : 
            Number of X two-qubit mixer gates to sample for the stochastic parameter shift.
        n_gamma_pair : 
            Number of two-qubit cost gates to sample for the stochastic parameter shift.
        n_gamma_single : 
            Number of single-qubit cost gates to sample for the stochastic parameter shift.

    RETURNS
    -------
    grad_sps_func:
        Callable derivative function.
    """
    
    n_beta_single = gradient_options['n_beta_single']
    n_beta_pair = gradient_options['n_beta_pair']
    n_gamma_single = gradient_options['n_gamma_single']
    n_gamma_pair = gradient_options['n_gamma_pair']
    
    beta_single_len = len(params_ext.betas_singles[0])
    beta_pair_len = len(params_ext.betas_pairs[0])
    gamma_single_len = len(params_ext.gammas_singles[0])
    gamma_pair_len = len(params_ext.gammas_pairs[0])
    
    coeffs_list = params_std.p*params_std.mixer_1q_coeffs + params_std.p*params_std.mixer_2q_coeffs + \
            params_std.p*params_std.cost_1q_coeffs + params_std.p*params_std.cost_2q_coeffs
    
    assert (-1 <= n_beta_single <= beta_single_len) and (-1 <= n_beta_pair <= beta_pair_len) and (-1 <= n_gamma_single <= gamma_single_len) and (-1 <= n_gamma_pair <= gamma_pair_len),\
        f"Invalid (n_beta_single, n_beta_pair, n_gamma_pair, n_gamma_single). Each n must be -1 or integers less than or equals to ({str(beta_single_len)}, {str(beta_pair_len)}, {str(gamma_single_len)}, {str(gamma_pair_len)})"

    if n_beta_single == -1: n_beta_single = beta_single_len
    if n_beta_pair == -1: n_beta_pair = beta_pair_len
    if n_gamma_single == -1: n_gamma_single = gamma_single_len
    if n_gamma_pair == -1: n_gamma_pair = gamma_pair_len
        
    fun = update_and_compute_expectation(backend_obj, params_ext, logger)
    
    def grad_sps_func(args):

        # Convert standard to extended parameters before applying parameter shift
        args_ext = params_std.convert_to_ext(args)
        
        grad_ext = np.zeros(len(args_ext))

        # Generate lists of random gates to sample. Note : Gates sampled in each layer is not necessarily the same, but the number of sampled gates in each layer is the same.
        sampled_indices = []
        for p in range(params_std.p):
            sampled_indices.append(random.sample(range(p*len(params_std.mixer_1q_coeffs) , (p+1)*len(params_std.mixer_1q_coeffs)), n_beta_single))

            sampled_indices.append(random.sample(range(params_std.p*len(params_std.mixer_1q_coeffs) + p*len(params_std.mixer_2q_coeffs) , params_std.p*len(params_std.mixer_1q_coeffs) + (p+1)*len(params_std.mixer_2q_coeffs)), n_beta_pair))

            sampled_indices.append(random.sample(range(params_std.p*(len(params_std.mixer_1q_coeffs) + len(params_std.mixer_2q_coeffs)) + p*len(params_std.cost_1q_coeffs) , params_std.p*(len(params_std.mixer_1q_coeffs) + len(params_std.mixer_2q_coeffs)) + (p+1)*len(params_std.cost_1q_coeffs)), n_gamma_single))

            sampled_indices.append(random.sample(range(params_std.p*(len(params_std.mixer_1q_coeffs) + len(params_std.mixer_2q_coeffs) + len(params_std.cost_1q_coeffs)) + p*len(params_std.cost_2q_coeffs) , params_std.p*(len(params_std.mixer_1q_coeffs) + len(params_std.mixer_2q_coeffs) + len(params_std.cost_1q_coeffs)) + (p+1)*len(params_std.cost_2q_coeffs)), n_gamma_pair))
    
        sampled_indices = [item for sublist in sampled_indices for item in sublist]

        # Apply parameter shifts
        for i in range(len(args_ext)):
            if (i) in sampled_indices:
                vect_eta = np.zeros(len(args_ext))
                vect_eta[i] = 1
                r = coeffs_list[i]
                grad_ext[i] = r*(fun(args_ext + (np.pi/(4*r))*vect_eta) -
                                 fun(args_ext - (np.pi/(4*r))*vect_eta))

        # Convert extended param. gradient form back into std param. form
        
        m1q_entries = grad_ext[:params_std.p*len(params_std.mixer_1q_coeffs)]
        m2q_entries = grad_ext[params_std.p*len(params_std.mixer_1q_coeffs) : params_std.p*len(params_std.mixer_2q_coeffs)]
        c1q_entries = grad_ext[params_std.p*len(params_std.mixer_1q_coeffs) + params_std.p*len(params_std.mixer_2q_coeffs): params_std.p*len(params_std.mixer_1q_coeffs) + params_std.p*len(params_std.mixer_2q_coeffs) + params_std.p*len(params_std.cost_1q_coeffs)]
        c2q_entries = grad_ext[params_std.p*len(params_std.mixer_1q_coeffs) + params_std.p*len(params_std.mixer_2q_coeffs) + params_std.p*len(params_std.cost_1q_coeffs):]
        subdivided_ext_grad = [m1q_entries, m2q_entries, c1q_entries, c2q_entries]
        
        # Sum up gradients (due to the chain rule), and re-express in standard form.
        mat = np.zeros((4, params_std.p))
        for i in range(4):  # 4 types of terms
            for j in range(params_std.p):
                mat[i][j] = np.sum(subdivided_ext_grad[i][j*int(len(subdivided_ext_grad[i]) /
                                   params_std.p):(j+1)*int(len(subdivided_ext_grad[i])/params_std.p)])

        grad_std = list(np.sum(mat[:2], axis=0)) + \
            list(np.sum(mat[2:], axis=0))
        
        return np.array(grad_std)

    return grad_sps_func


def hessian_fd(backend_obj, params, hessian_options, logger):
    """
    Returns a callable function that calculates the hessian with the finite difference method.

    PARAMETERS
    ----------
    backend_obj : `QAOABaseBackend`
        backend object that computes expectation values when executed.

    params : `QAOAVariationalBaseParams`
        variational parameters object.

    hessian_options :
        hessian_stepsize : 
            stepsize of finite difference.

    RETURNS
    -------
    hessian_fd_func:
        Callable derivative function.

    """

    eta = hessian_options['stepsize']
    fun = update_and_compute_expectation(backend_obj, params, logger)

    def hessian_fd_func(args):
        hess = np.zeros((len(args), len(args)))

        for i in range(len(args)):
            for j in range(len(args)):
                vect_eta1 = np.zeros(len(args))
                vect_eta2 = np.zeros(len(args))
                vect_eta1[i] = 1
                vect_eta2[j] = 1

                if i == j:
                    # Central diff. hessian diagonals (https://v8doc.sas.com/sashtml/ormp/chap5/sect28.htm)
                    hess[i][i] = (-fun(args+2*eta*vect_eta1) + 16*fun(args + eta*vect_eta1) - 30*fun(
                        args) + 16*fun(args-eta*vect_eta1)-fun(args-2*eta*vect_eta1))/(12*eta**2)
                #grad_diff[i] = (grad_fd_ext(params + (eta/2)*vect_eta)[i] - grad_fd_ext(params - (eta/2)*vect_eta)[i])/eta
                else:
                    hess[i][j] = (fun(args + eta*vect_eta1 + eta*vect_eta2)-fun(args + eta*vect_eta1 - eta*vect_eta2)-fun(
                        args - eta*vect_eta1 + eta*vect_eta2)+fun(args - eta*vect_eta1 - eta*vect_eta2))/(4*eta**2)

        return hess

    return hessian_fd_func


def grad_spsa(backend_obj, params, gradient_options, logger):
    """
    Returns a callable function that calculates the gradient approxmiation with the Simultaneous Perturbation Stochastic Approximation (SPSA) method.

    PARAMETERS
    ----------
    backend_obj : `QAOABaseBackend`
        backend object that computes expectation values when executed. 

    params : `QAOAVariationalBaseParams`
        variational parameters object.

    gradient_options : `dict`
        gradient_stepsize : 
            stepsize of stochastic shift.

    RETURNS
    -------
    grad_spsa_func: `Callable`
        Callable derivative function.

    """
    c = gradient_options['stepsize']
    fun = update_and_compute_expectation(backend_obj, params, logger)

    def grad_spsa_func(args):
        delta = (2*np.random.randint(0, 2, size=len(args))-1)
        return np.real((fun(args + c*delta) - fun(args - c*delta))*delta/(2*c))

    return grad_spsa_func



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os
from typing import Optional, List
# AWS Braket imports

from braket.circuits import Circuit
from braket.circuits.gates import H
from braket.circuits.result_types import Probability
from braket.circuits.free_parameter import FreeParameter
from braket.jobs.metrics import log_metric

from ...devices import DeviceAWS
from ...basebackend import QAOABaseBackendShotBased, QAOABaseBackendCloud, QAOABaseBackendParametric
from ...qaoa_parameters.baseparams import QAOACircuitParams, QAOAVariationalBaseParams

class QAOAAWSQPUBackend(QAOABaseBackendParametric, QAOABaseBackendCloud, QAOABaseBackendShotBased):
    """
    A QAOA simulator as well as for real QPU using Amazon Braket as backend

    Parameters
    ----------
    device: `DeviceAWS`
        An object of the class ``DeviceAWS`` which contains the credentials
        for accessing the QPU via cloud and the name of the device.
    circuit_params: `QAOACircuitParams`
        An object of the class ``QAOACircuitParams`` which contains information on 
        circuit construction and depth of the circuit.
    n_shots: `int`
        The number of shots to be taken for each circuit.
    prepend_state: `QuantumCircuit`
        The state prepended to the circuit.
    append_state: `QuantumCircuit`
        The state appended to the circuit.
    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the 
        QAOA part of the circuit.
    cvar_alpha: `float`
        The value of alpha for the CVaR method.
    disable_qubit_rewiring: `bool`
        A boolean that determines whether qubit routing on the provider's end is 
        used. This is False by default. Not all providers provide this feature.
    """

    def __init__(self,
                 circuit_params: QAOACircuitParams,
                 device: DeviceAWS,
                 n_shots: int,
                 prepend_state: Optional[Circuit],
                 append_state: Optional[Circuit],
                 init_hadamard: bool,
                 cvar_alpha: float,
                 qubit_layout: List[int] = [], 
                 disable_qubit_rewiring: bool = False):

        QAOABaseBackendShotBased.__init__(self,
                                          circuit_params,
                                          n_shots,
                                          prepend_state,
                                          append_state,
                                          init_hadamard,
                                          cvar_alpha)
        QAOABaseBackendCloud.__init__(self, device)

        self.qureg = self.circuit_params.qureg
        self.qubit_layout = self.qureg if qubit_layout == [] else qubit_layout
        self.disable_qubit_rewiring = disable_qubit_rewiring

        if self.prepend_state:
            assert self.n_qubits >= len(prepend_state.qubits), "Cannot attach a bigger circuit" \
                                                               " to the QAOA routine"

        if self.device.provider_connected and self.device.qpu_connected:
            self.backend_qpu = self.device.backend_device
        elif self.device.provider_connected is True and self.device.qpu_connected is False:
            raise Exception(
                'Connection to AWS was made. Error connecting to the specified backend.')
        elif self.device.provider_connected is True and self.device.qpu_connected is None:
            raise Exception(
                'Connection to AWS was made. A device name was not specified.')
        else:
            raise Exception('Error connecting to AWS.')
            
        self.parametric_circuit = self.parametric_qaoa_circuit()

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> Circuit:
        """
        The final QAOA circuit to be executed on the QPU.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        qaoa_circuit: `Circuit`
            The final QAOA circuit constructed using the angles from variational params.
        """
        
        angles_list = self.obtain_angles_for_pauli_list(
            self.abstract_circuit, params)
        memory_map = dict(zip([each_free_param_obj.name for each_free_param_obj in self.braket_parameter_list], angles_list))
        new_parametric_circuit = self.parametric_circuit.make_bound_circuit(
            memory_map)
        return new_parametric_circuit

    @property
    def parametric_qaoa_circuit(self) -> Circuit:
        """
        Creates a parametric QAOA circuit, given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit.
        """ 
        parametric_circuit = Circuit()
        if self.prepend_state:
            parametric_circuit += self.prepend_state
            
        # Initial state is all |+>
        if self.init_hadamard:
            for each_qubit in self.qubit_layout:
                parametric_circuit += H.h(each_qubit)

        self.braket_parameter_list = []
        for each_gate in self.abstract_circuit:
            angle_param = FreeParameter(str(each_gate.pauli_label))
            self.braket_parameter_list.append(angle_param)
            each_gate.rotation_angle = angle_param
            decomposition = each_gate.decomposition('standard')
            # using the list above, construct the circuit
            for each_tuple in decomposition:
                gate = each_tuple[0]()
                parametric_circuit = gate.apply_braket_gate(*each_tuple[1], parametric_circuit)

        if self.append_state:
            parametric_circuit += self.append_state
            
        parametric_circuit += Probability.probability()

        return parametric_circuit

    def get_counts(self, params: QAOAVariationalBaseParams) -> dict:
        """
        Execute the circuit and obtain the counts

        Parameters
        ----------
        params: QAOAVariationalBaseParams
            The QAOA parameters - an object of one of the parameter classes, containing 
            variable parameters.

        Returns
        -------
            A dictionary with the bitstring as the key and the number of counts 
            as its value.
        """

        circuit = self.qaoa_circuit(params)

        job_state = False
        no_of_job_retries = 0
        max_job_retries = 5

        while job_state == False:
            job = self.backend_qpu.run(circuit, 
                                       (self.device.s3_bucket_name, 
                                        self.device.folder_name), 
                                       shots = self.n_shots, 
                                       disable_qubit_rewiring = self.disable_qubit_rewiring)
            
            try:
                self.job_id = job.id
                
                job_result = job.result()
                
                # If there was an issue with the job sent, send again.
                if job.state() in ['FAILED', 'CANCELLED'] or job_result == None:
                    raise ValueError
                    
                counts = job_result.measurement_counts
                    
            except ValueError:
                print('The task has failed or was cancelled by AWS. Resending task.')
                no_of_job_retries += 1
                    
            except Exception as e:
                print(e, '\n')
                print("An unknown error occurred while trying to retrieve task results. Resending task.")
                no_of_job_retries += 1
                
            else:
                job_state = True
                
            finally:
                if no_of_job_retries >= max_job_retries:
                    raise ConnectionError(
                        "An Error Occurred with the Task(s) sent to AWS.")

        # Expose counts
        self.measurement_outcomes = counts
        return counts
    
    def log_with_backend(self, metric_name: str, value, iteration_number) -> None:
        
        """
        If using AWS Jobs, these values will be logged.
        """
        
        try:
            if os.environ["AMZN_BRAKET_JOB_NAME"] is not None:
                in_jobs = True
        except KeyError:
            in_jobs = False
        
        if in_jobs:
            log_metric(
                metric_name=metric_name,
                value=value,
                iteration_number=iteration_number,
            )

    def circuit_to_qasm(self, params: QAOAVariationalBaseParams) -> str:
        """
        A method to convert the entire QAOA `QuantumCircuit` object into 
        a OpenQASM string
        """
        raise NotImplementedError()
        # qasm_string = self.qaoa_circuit(params).qasm(formatted=True)
        # return qasm_string

    def reset_circuit(self):
        """
        Reset self.circuit after performing a computation
        """
        raise NotImplementedError()



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC
import numpy as np
import copy
from openqaoa.devices import DeviceLocal, DeviceBase

from openqaoa.problems.problem import QUBO
from openqaoa.problems.helper_functions import convert2serialize
from openqaoa.workflows.parameters.qaoa_parameters import CircuitProperties, BackendProperties, ClassicalOptimizer
from openqaoa.workflows.parameters.rqaoa_parameters import RqaoaParameters, ALLOWED_RQAOA_TYPES
from openqaoa.qaoa_parameters import Hamiltonian, QAOACircuitParams, create_qaoa_variational_params
from openqaoa.utilities import get_mixer_hamiltonian, ground_state_hamiltonian, exp_val_hamiltonian_termwise
from openqaoa.backends.qaoa_backend import get_qaoa_backend, DEVICE_NAME_TO_OBJECT_MAPPER, DEVICE_ACCESS_OBJECT_MAPPER
from openqaoa.optimizers.qaoa_optimizer import get_optimizer
from openqaoa.basebackend import QAOABaseBackendStatevector
from openqaoa import rqaoa
from openqaoa.rqaoa.rqaoa_results import RQAOAResults


class Optimizer(ABC):
    """
    Abstract class to represent an optimizer

    It's basic usage consists of 

     #. Initialization
     #. Compilation
     #. Optimization
    
    Attributes
    ----------
        device: `DeviceBase`
            Device to be used by the optimizer
        backend_properties: `BackendProperties`
            The backend properties of the optimizer workflow. Use to set the backend properties such as the number of shots and the cvar values.
            For a complete list of its parameters and usage please see the method set_backend_properties
        classical_optimizer: `ClassicalOptimizer`
            The classical optimiser properties of the optimizer workflow. Use to set the classical optimiser needed for the classical optimisation part of the optimizer routine.
            For a complete list of its parameters and usage please see the method set_classical_optimizer
        local_simulators: `list[str]`
            A list containing the available local simulators
        cloud_provider: `list[str]`
            A list containing the available cloud providers
        compiled: `Bool`
            A boolean flag to check whether the optimizer object has been correctly compiled at least once     
    """

    def __init__(self, device=DeviceLocal('vectorized')):  
        """
        Initialize the optimizer class.

        Parameters
        ----------
        device: `DeviceBase`
            Device to be used by the optimizer. Default is using the local 'vectorized' simulator.
        """
        
        self.device = device
        self.backend_properties = BackendProperties()
        self.classical_optimizer = ClassicalOptimizer()
        self.local_simulators = list(DEVICE_NAME_TO_OBJECT_MAPPER.keys())
        self.cloud_provider = list(DEVICE_ACCESS_OBJECT_MAPPER.keys())
        self.compiled = False

    def asdict(self):
        attributes_dict = convert2serialize(self)
        return attributes_dict

    def set_device(self, device: DeviceBase):
        """"
        Specify the device to be used by the QAOA.

        Parameters
        ----------
        location: `str`
            Can be either local, qcs, or ibmq
        name: `str`
            The name of the device to be used, for local simulators please refer to `q.local_simulators`.
            For cloud providers please refer to the provider's naming conventions
        """
        self.device = device

    def set_backend_properties(self, **kwargs):
        """
        Set the backend properties

        Parameters
        -------------------
            device: DeviceBase
            prepend_state: [Union[QuantumCircuitBase,List[complex], np.ndarray]
                The state prepended to the circuit.
            append_state: [Union[QuantumCircuitBase,List[complex], np.ndarray]
                The state prepended to the circuit.
            init_hadamard: bool
            Whether to apply a Hadamard gate to the beginning of the 
                QAOA part of the circuit.. Defaults to `True`
            n_shots: int
            Optional argument to specify the number of shots required to run QAOA computations
                on shot-based simulators and QPUs. Defaults to 100.
            cvar_alpha: float
                The value of alpha for the CVaR cost function
            noise_model: `qiskit.providers.aer.noise.NoiseModel`
                    The Qiskit noise model to be used for the simulation.
            qiskit_simulation_method: str, optional
                The method to be used for the simulation.
            seed_simulator: int
                Optional argument to initialize a pseudorandom solution. Default None
            active_reset:
                #TODO
            rewiring:
                Rewiring scheme to be used for Pyquil. 
                Either 'PRAGMA INITIAL_REWIRING "NAIVE"' or 
                'PRAGMA INITIAL_REWIRING "PARTIAL"'. If None, defaults to NAIVE
            disable_qubit_rewiring: `bool`
                Disable automatic qubit rewiring on AWS braket backend
        """

        for key, value in kwargs.items():
            if hasattr(self.backend_properties, key):
                pass# setattr(self.backend_properties, key, value)
            else:
                raise ValueError(
                    f'Specified argument `{value}` for `{key}` in set_backend_properties is not supported')

        self.backend_properties = BackendProperties(**kwargs)
        return None

    def set_classical_optimizer(self, **kwargs):
        """
        Set the parameters for the classical optimizer to be used in the optimizers workflow

        Parameters
        ----------
            method: str
                The classical optimization method. Choose from:
                 ['imfil','bobyqa','snobfit']
                 ['vgd', 'sgd', 'rmsprop'] 
                 ['nelder-mead','powell','cg','bfgs','newton-cg','l-bfgs-b','cobyla'] 
            maxiter : Optional[int]
                Maximum number of iterations.
            maxfev : Optional[int]
                Maximum number of function evaluations.
            jac: str
                Method to compute the gradient vector. Choose from:
                    - ['finite_difference', 'param_shift', 'stoch_param_shift', 'grad_spsa']        
            hess: str
                Method to compute the hessian. Choose from:
                    - ['finite_difference', 'param_shift', 'stoch_param_shift', 'grad_spsa']
            constraints: scipy.optimize.LinearConstraints, scipy.optimize.NonlinearConstraints  
                Scipy-based constraints on parameters of optimization. Will be available soon
            bounds: scipy.optimize.Bounds
                Scipy-based bounds on parameters of optimization. Will be available soon
            tol : float
                Tolerance before the optimizer terminates; if `tol` is larger than
                the difference between two steps, terminate optimization.
            optimizer_options : dict
                Dictionary of optimiser-specific arguments.
                    stepsize : float
                        Step size of each gradient descent step.
                    decay : float
                        Stepsize decay parameter of RMSProp.
                    eps : float
                        Small number to prevent division by zero for RMSProp.
                    lambd : float
                        Small number to prevent singularity of QFIM matrix for Natural Gradient Descent.
            ramp_time: float
                The slope(rate) of linear ramp initialisation of QAOA parameters.
            jac_options : dict
                Dictionary that specifies gradient-computation options according to method chosen in 'jac'.
            hess_options : dict
                Dictionary that specifies Hessian-computation options according to method chosen in 'hess'.
            save_intermediate: bool
                If True, the intermediate parameters of the optimization and job ids, if available, are saved throughout the run. This is set to False by default.
        """
        for key, value in kwargs.items():
            if hasattr(self.classical_optimizer, key):
                pass #setattr(self.classical_optimizer, key, value)
            else:
                raise ValueError(
                    'Specified argument is not supported by the Classical Optimizer')

        self.classical_optimizer = ClassicalOptimizer(**kwargs)
        return None

    def compile():
        raise NotImplementedError

    def optimize():
        raise NotImplementedError


class QAOA(Optimizer):
    """
    A class implementing a QAOA workflow end to end.

    It's basic usage consists of 
    1. Initialization
    2. Compilation
    3. Optimization

    .. note::
        The attributes of the QAOA class should be initialized using the set methods of QAOA. For example, to set the circuit's depth to 10 you should run `set_circuit_properties(p=10)`

    Attributes
    ----------
        device: `DeviceBase`
            Device to be used by the optimizer
        circuit_properties: `CircuitProperties`
            The circuit properties of the QAOA workflow. Use to set depth `p`, choice of parametrisation, parameter initialisation strategies, mixer hamiltonians.
            For a complete list of its parameters and usage please see the method set_circuit_properties
        backend_properties: `BackendProperties`
            The backend properties of the QAOA workflow. Use to set the backend properties such as the number of shots and the cvar values.
            For a complete list of its parameters and usage please see the method set_backend_properties
        classical_optimizer: `ClassicalOptimizer`
            The classical optimiser properties of the QAOA workflow. Use to set the classical optimiser needed for the classical optimisation part of the QAOA routine.
            For a complete list of its parameters and usage please see the method set_classical_optimizer
        local_simulators: `list[str]`
            A list containing the available local simulators
        cloud_provider: `list[str]`
            A list containing the available cloud providers
        mixer_hamil: Hamiltonian
            The desired mixer hamiltonian
        cost_hamil: Hamiltonian
            The desired mixer hamiltonian
        circuit_params: QAOACircuitParams
            the abstract and backend-agnostic representation of the underlying QAOA parameters
        variate_params: QAOAVariationalBaseParams
            The variational parameters. These are the parameters to be optimised by the classical optimiser
        backend: VQABaseBackend
            The openQAOA representation of the backend to be used to execute the quantum circuit
        optimizer: OptimizeVQA
            The classical optimiser
        results: `Result`
            Contains the logs of the optimisation process
        compiled: `Bool`
            A boolean flag to check whether the QAOA object has been correctly compiled at least once

    Examples
    --------
    Examples should be written in doctest format, and should illustrate how
    to use the function.

    >>> q = QAOA()
    >>> q.compile(QUBO)
    >>> q.optimise()

    Where `QUBO` is a an instance of `openqaoa.problems.problem.QUBO`

    If you want to use non-default parameters:

    >>> q_custom = QAOA()
    >>> q_custom.set_circuit_properties(p=10, param_type='extended', init_type='ramp', mixer_hamiltonian='x')
    >>> q_custom.set_device_properties(device_location='qcs', device_name='Aspen-11', cloud_credentials={'name' : "Aspen11", 'as_qvm':True, 'execution_timeout' : 10, 'compiler_timeout':10})
    >>> q_custom.set_backend_properties(n_shots=200, cvar_alpha=1)
    >>> q_custom.set_classical_optimizer(method='nelder-mead', maxiter=2)
    >>> q_custom.compile(qubo_problem)
    >>> q_custom.optimize()
    """

    def __init__(self, device=DeviceLocal('vectorized')):
        """
        Initialize the QAOA class.

        Parameters
        ----------
            device: `DeviceBase`
                Device to be used by the optimizer. Default is using the local 'vectorized' simulator.
        """
        super().__init__(device)
        self.circuit_properties = CircuitProperties()

    def set_circuit_properties(self, **kwargs):
        """
        Specify the circuit properties to construct QAOA circuit

        Parameters
        -------------------
            qubit_register: `list`
                Select the desired qubits to run the QAOA program. Meant to be used as a qubit
                selector for qubits on a QPU. Defaults to a list from 0 to n-1 (n = number of qubits)
            p: `int`
                Depth `p` of the QAOA circuit
            q: `int`
                Analogue of `p` of the QAOA circuit in the Fourier parameterisation
            param_type: `str`
                Choose the QAOA circuit parameterisation. Currently supported parameterisations include:
                `'standard'`: Standard QAOA parameterisation
                `'standard_w_bias'`: Standard QAOA parameterisation with a separate parameter for single-qubit terms.
                `'extended'`: Individual parameter for each qubit and each term in the Hamiltonian.
                `'fourier'`: Fourier circuit parameterisation
                `'fourier_extended'`: Fourier circuit parameterisation with individual parameter for each qubit and term in Hamiltonian.
                `'fourier_w_bias'`: Fourier circuit parameterisation with aseparate parameter for single-qubit terms
            init_type: `str`
                Initialisation strategy for the QAOA circuit parameters. Allowed init_types:
                `'rand'`: Randomly initialise circuit parameters
                `'ramp'`: Linear ramp from Hamiltonian initialisation of circuit parameters (inspired from Quantum Annealing)
                `'custom'`: User specified initial circuit parameters
            mixer_hamiltonian: `str`
                Parameterisation of the mixer hamiltonian:
                `'x'`: Randomly initialise circuit parameters
                `'xy'`: Linear ramp from Hamiltonian initialisation of circuit 
            mixer_qubit_connectivity: `[Union[List[list],List[tuple], str]]`
                The connectivity of the qubits in the mixer Hamiltonian. Use only if `mixer_hamiltonian = xy`. The user can specify the 
                connectivity as a list of lists, a list of tuples, or a string chosen from ['full', 'chain', 'star'].
            mixer_coeffs: `list`
                The coefficients of the mixer Hamiltonian. By default all set to -1
            annealing_time: `float`
                Total time to run the QAOA program in the Annealing parameterisation (digitised annealing)
            linear_ramp_time: `float`
                The slope(rate) of linear ramp initialisation of QAOA parameters.
            variational_params_dict: `dict`
                Dictionary object specifying the initial value of each circuit parameter for the chosen parameterisation, if the `init_type` is selected as `'custom'`.    
                For example, for standard parametrisation set {'betas': [0.1, 0.2, 0.3], 'gammas': [0.1, 0.2, 0.3]}
        """

        for key, value in kwargs.items():
            if hasattr(self.circuit_properties, key):
                pass
            else:
                raise ValueError(
                    "Specified argument is not supported by the circuit")
        self.circuit_properties = CircuitProperties(**kwargs)

        return None

    def compile(self, problem: QUBO = None, verbose: bool = False):
        """
        Initialise the trainable parameters for QAOA according to the specified
        strategies and by passing the problem statement

        .. note::
            Compilation is necessary because it is the moment where the problem statement and the QAOA instructions are used to build the actual QAOA circuit.

        .. tip::
            Set Verbose to false if you are running batch computations! 

        Parameters
        ----------
        problem: `Problem`
            QUBO problem to be solved by QAOA
        verbose: bool
            Set True to have a summary of QAOA to displayed after compilation
        """

        assert isinstance(problem, QUBO), "The problem must be converted into QUBO form"
        
        self.cost_hamil = Hamiltonian.classical_hamiltonian(
            terms=problem.terms, coeffs=problem.weights, constant=problem.constant)
        
        self.mixer_hamil = get_mixer_hamiltonian(n_qubits=self.cost_hamil.n_qubits,
                                                 mixer_type=self.circuit_properties.mixer_hamiltonian,
                                                 qubit_connectivity=self.circuit_properties.mixer_qubit_connectivity,
                                                 coeffs=self.circuit_properties.mixer_coeffs)

        self.circuit_params = QAOACircuitParams(
            self.cost_hamil, self.mixer_hamil, p=self.circuit_properties.p)
        self.variate_params = create_qaoa_variational_params(qaoa_circuit_params=self.circuit_params,
                                                             params_type=self.circuit_properties.param_type,
                                                             init_type=self.circuit_properties.init_type, 
                                                             variational_params_dict=self.circuit_properties.variational_params_dict,
                                                             linear_ramp_time=self.circuit_properties.linear_ramp_time, 
                                                             q=self.circuit_properties.q, 
                                                             seed=self.circuit_properties.seed,
                                                             total_annealing_time=self.circuit_properties.annealing_time)

        self.backend = get_qaoa_backend(circuit_params=self.circuit_params,
                                        device=self.device,
                                        **self.backend_properties.__dict__)
        self.optimizer = get_optimizer(vqa_object=self.backend,
                                       variational_params=self.variate_params,
                                       optimizer_dict=self.classical_optimizer.asdict())

        self.compiled = True
        
        if verbose:
            print('\t \033[1m ### Summary ###\033[0m')
            print(f'OpenQAOA has been compiled with the following properties')
            print(
                f'Solving QAOA with \033[1m {self.device.device_name} \033[0m on  \033[1m{self.device.device_location}\033[0m')
            print(f'Using p={self.circuit_properties.p} with {self.circuit_properties.param_type} parameters initialized as {self.circuit_properties.init_type}')

            if self.device.device_name == 'vectorized':
                print(
                    f'OpenQAOA will optimize using \033[1m{self.classical_optimizer.method}\033[0m, with up to \033[1m{self.classical_optimizer.maxiter}\033[0m maximum iterations')

            else:
                print(
                    f'OpenQAOA will optimize using \033[1m{self.classical_optimizer.method}\033[0m, with up to \033[1m{self.classical_optimizer.maxiter}\033[0m maximum iterations. Each iteration will contain \033[1m{self.backend_properties.n_shots} shots\033[0m')
                print(
                    f'The total number of shots is set to maxiter*shots = {self.classical_optimizer.maxiter*self.backend_properties.n_shots}')

        return None

    def optimize(self, verbose=False):
        '''
        A method running the classical optimisation loop
        '''

        if self.compiled == False:
            raise ValueError('Please compile the QAOA before optimizing it!')

        self.optimizer.optimize()
        # TODO: results and qaoa_results will differ
        self.results = self.optimizer.qaoa_result

        if verbose:
            print(f'optimization completed.')
        return


class RQAOA(Optimizer):
    """
    A class implementing a RQAOA workflow end to end.

    It's basic usage consists of 
    1. Initialization
    2. Compilation
    3. Optimization

    .. note::
        The attributes of the RQAOA class should be initialized using the set methods of QAOA. For example, to set the qaoa circuit's depth to 10 you should run `set_circuit_properties(p=10)`

    Attributes
    ----------
        device: `DeviceBase`
            Device to be used by the optimizer
        backend_properties: `BackendProperties`
            The backend properties of the RQAOA workflow. These properties will be used to run QAOA at each RQAOA step.
            Use to set the backend properties such as the number of shots and the cvar values.
            For a complete list of its parameters and usage please see the method set_backend_properties
        classical_optimizer: `ClassicalOptimizer`
            The classical optimiser properties of the RQAOA workflow. 
            Use to set the classical optimiser needed for the classical optimisation part of the QAOA routine.
            For a complete list of its parameters and usage please see the method set_classical_optimizer
        local_simulators: `list[str]`
            A list containing the available local simulators
        cloud_provider: `list[str]`
            A list containing the available cloud providers
        compiled: `Bool`
            A boolean flag to check whether the optimizer object has been correctly compiled at least once
        circuit_properties: `CircuitProperties`
            The circuit properties of the RQAOA workflow. These properties will be used to run QAOA at each RQAOA step.
            Use to set depth `p`, choice of parametrisation, parameter initialisation strategies, mixer hamiltonians.
            For a complete list of its parameters and usage please see the method set_circuit_properties
        rqaoa_parameters: `RqaoaParameters`
            Set of parameters containing all the relevant information for the recursive procedure of RQAOA.
        results: `RQAOAResults`
            The results of the RQAOA optimization. 
            Dictionary containing all the information about the RQAOA run: the
            solution states and energies (key: 'solution'), the output of the classical 
            solver (key: 'classical_output'), the elimination rules for each step
            (key: 'elimination_rules'), the number of eliminations at each step (key: 'schedule'), 
            total number of steps (key: 'number_steps'), the intermediate QUBO problems and the 
            intermediate QAOA objects that have been optimized in each RQAOA step (key: 'intermediate_problems').
            This object (`RQAOAResults`) is a dictionary with some custom methods as RQAOAResults.get_hamiltonian_step(i) 
            which get the hamiltonian of reduced problem of the i-th step. To see the full list of methods please see the
            RQAOAResults class.  

    Examples
    --------
    Examples should be written in doctest format, and should illustrate how
    to use the function.

    >>> r = RQAOA()
    >>> r.compile(QUBO)
    >>> r.optimise()

    Where `QUBO` is a an instance of `openqaoa.problems.problem.QUBO`

    If you want to use non-default parameters:

    Standard/custom (default) type:
    >>> r = QAOA()
    >>> r.set_circuit_properties(p=10, param_type='extended', init_type='ramp', mixer_hamiltonian='x')
    >>> r.set_device_properties(device_location='qcs', device_name='Aspen-11', cloud_credentials={'name' : "Aspen11", 'as_qvm':True, 'execution_timeout' : 10, 'compiler_timeout':10})
    >>> r.set_backend_properties(n_shots=200, cvar_alpha=1)
    >>> r.set_classical_optimizer(method='nelder-mead', maxiter=2)
    >>> r.set_rqaoa_parameters(n_cutoff = 5, steps=[1,2,3,4,5])
    >>> r.compile(qubo_problem)
    >>> r.optimize()

    Ada-RQAOA:
    >>> r_adaptive = QAOA()
    >>> r_adaptive.set_circuit_properties(p=10, param_type='extended', init_type='ramp', mixer_hamiltonian='x')
    >>> r_adaptive.set_device_properties(device_location='qcs', device_name='Aspen-11', cloud_credentials={'name' : "Aspen11", 'as_qvm':True, 'execution_timeout' : 10, 'compiler_timeout':10})
    >>> r_adaptive.set_backend_properties(n_shots=200, cvar_alpha=1)
    >>> r_adaptive.set_classical_optimizer(method='nelder-mead', maxiter=2)
    >>> r_adaptive.set_rqaoa_parameters(rqaoa_type = 'adaptive', n_cutoff = 5, n_max=5)
    >>> r_adaptive.compile(qubo_problem)
    >>> r_adaptive.optimize()
    """

    def __init__(self, device: DeviceBase=DeviceLocal('vectorized')):
        """
        Initialize the RQAOA class.

        Parameters
        ----------
            device: `DeviceBase`
                Device to be used by the optimizer. Default is using the local 'vectorized' simulator.
        """
        super().__init__(device) # use the parent class to initialize 
        self.circuit_properties = CircuitProperties()
        self.rqaoa_parameters = RqaoaParameters()

        # varaible that will store results object (when optimize is called)
        self.results = RQAOAResults()

    def set_circuit_properties(self, **kwargs): 
        """
        Specify the circuit properties to construct the QAOA circuits

        Parameters
        ----------
            qubit_register: `list`
                Select the desired qubits to run the QAOA program. Meant to be used as a qubit
                selector for qubits on a QPU. Defaults to a list from 0 to n-1 (n = number of qubits)
            p: `int`
                Depth `p` of the QAOA circuit
            q: `int`
                Analogue of `p` of the QAOA circuit in the Fourier parameterisation
            param_type: `str`
                Choose the QAOA circuit parameterisation. Currently supported parameterisations include:
                `'standard'`: Standard QAOA parameterisation
                `'standard_w_bias'`: Standard QAOA parameterisation with a separate parameter for single-qubit terms.
                `'extended'`: Individual parameter for each qubit and each term in the Hamiltonian.
                `'fourier'`: Fourier circuit parameterisation
                `'fourier_extended'`: Fourier circuit parameterisation with individual parameter for each qubit and term in Hamiltonian.
                `'fourier_w_bias'`: Fourier circuit parameterisation with aseparate parameter for single-qubit terms
            init_type: `str`
                Initialisation strategy for the QAOA circuit parameters. Allowed init_types:
                `'rand'`: Randomly initialise circuit parameters
                `'ramp'`: Linear ramp from Hamiltonian initialisation of circuit parameters (inspired from Quantum Annealing)
                `'custom'`: User specified initial circuit parameters
            mixer_hamiltonian: `str`
                Parameterisation of the mixer hamiltonian:
                `'x'`: Randomly initialise circuit parameters
                `'xy'`: Linear ramp from Hamiltonian initialisation of circuit 
            mixer_qubit_connectivity: `[Union[List[list],List[tuple], str]]`
                The connectivity of the qubits in the mixer Hamiltonian. Use only if `mixer_hamiltonian = xy`. The user can specify the 
                connectivity as a list of lists, a list of tuples, or a string chosen from ['full', 'chain', 'star'].
            mixer_coeffs: `list`
                The coefficients of the mixer Hamiltonian. By default all set to -1
            annealing_time: `float`
                Total time to run the QAOA program in the Annealing parameterisation (digitised annealing)
            linear_ramp_time: `float`
                The slope(rate) of linear ramp initialisation of QAOA parameters.
            variational_params_dict: `dict`
                Dictionary object specifying the initial value of each circuit parameter for the chosen parameterisation, if the `init_type` is selected as `'custom'`.    
                For example, for standard parametrisation set {'betas': [0.1, 0.2, 0.3], 'gammas': [0.1, 0.2, 0.3]}
        """

        for key in kwargs.keys():
            if hasattr(self.circuit_properties, key):
                pass
            else:
                raise ValueError(
                    f"Specified argument {key} is not supported by the circuit")

        self.circuit_properties = CircuitProperties(**kwargs) 

        return None

    def set_rqaoa_parameters(self, **kwargs):
        """
        Specify the parameters to run a desired RQAOA program.

        Parameters
        ----------
        rqaoa_type: `int`
            String specifying the RQAOA scheme under which eliminations are computed. The two methods are 'custom' and
            'adaptive'. Defaults to 'custom'.
        n_max: `int`
            Maximum number of eliminations allowed at each step when using the adaptive method.
        steps: `Union[list,int]`
            Elimination schedule for the RQAOA algorithm. If an integer is passed, it sets the number of spins eliminated
            at each step. If a list is passed, the algorithm will follow the list to select how many spins to eliminate 
            at each step. Note that the list needs enough elements to specify eliminations from the initial number of qubits
            up to the cutoff value. If the list contains more, the algorithm will follow instructions until the cutoff value 
            is reached.
        n_cutoff: `int`
            Cutoff value at which the RQAOA algorithm obtains the solution classically.
        original_hamiltonian: `Hamiltonian`
            Hamiltonian encoding the original problem fed into the RQAOA algorithm.
        counter: `int`
            Variable to count the step in the schedule. If counter = 3 the next step is schedule[3]. 
            Default is 0, but can be changed to start in the position of the schedule that one wants.
        """

        for key in kwargs.keys():
            if hasattr(self.rqaoa_parameters, key):
                pass
            else:
                raise ValueError(
                    f'Specified argument {key} is not supported by RQAOA')

        self.rqaoa_parameters = RqaoaParameters(**kwargs) 

        return None

    def compile(self, problem: QUBO = None, verbose: bool = False):
        """
        Create a QAOA object and initialize it with the circuit properties, device, classical optimizer and
        backend properties specified by the user.
        This QAOA object will be used to run QAOA changing the problem to sove at each RQAOA step. 
        Here, the QAOA is compiled passing the problem statement, so to check that the compliation of 
        QAOA is correct. See the QAOA class.

        .. note::
            Compilation is necessary because it is the moment where the problem statement and the QAOA instructions are used to build the actual QAOA circuit.

        Parameters
        ----------
        problem: `Problem`
            QUBO problem to be solved by RQAOA
        verbose: bool
            !NotYetImplemented! Set true to have a summary of QAOA first step to displayed after compilation
        """

        # save the original problem
        self.problem = problem 

        # if type is custom and steps is an int, set steps correctly
        if self.rqaoa_parameters.rqaoa_type == "custom" and self.rqaoa_parameters.n_cutoff<=problem.n:

            n_cutoff = self.rqaoa_parameters.n_cutoff
            n_qubits = problem.n
            counter  = self.rqaoa_parameters.counter

            # If schedule for custom RQAOA is not given, we create a schedule such that 
            # n = self.rqaoa_parameters.steps spins is eliminated at a time
            if type(self.rqaoa_parameters.steps) is int:
                self.rqaoa_parameters.steps = [self.rqaoa_parameters.steps]*(n_qubits-n_cutoff)
            
            # In case a schedule is given, ensure there are enough steps in the schedule
            assert np.abs(n_qubits - n_cutoff - counter) <= sum(self.rqaoa_parameters.steps),\
                f"Schedule is incomplete, add {np.abs(n_qubits - n_cutoff - counter) - sum(self.rqaoa_parameters.steps)} more eliminations"


        # Create the qaoa object with the properties
        self._q = QAOA(self.device)
        self._q.circuit_properties  = self.circuit_properties
        self._q.backend_properties  = self.backend_properties
        self._q.classical_optimizer = self.classical_optimizer

        # compile qaoa object
        self._q.compile(problem, verbose=verbose)

        self.compiled = True

        return 

    def optimize(self, verbose=False):
        """
        Performs optimization using RQAOA with the `custom` method or the `adaptive` method.
        The elimination RQAOA loop will occur until the number of qubits is equal to the number of qubits specified in `n_cutoff`.
        In each loop, the QAOA will be run, then the eliminations will be computed, a new problem will be redefined
        and the QAOA will be recompiled with the new problem.
        Once the loop is complete, the final problem will be solved classically and the final solution will be reconstructed.
        Results will be stored in the `results` attribute.
        """

        # lists to append the eliminations and the qaoa objects
        elimination_tracker = []
        qaoa_steps = []
        problem_steps = []

        # get variables
        problem = self.problem  
        n_cutoff = self.rqaoa_parameters.n_cutoff
        n_qubits = problem.n
        counter = self.rqaoa_parameters.counter

        # copy the original qaoa object
        q = copy.deepcopy(self._q)

        # create a different max_terms function for each type 
        if self.rqaoa_parameters.rqaoa_type == "adaptive":
            f_max_terms = rqaoa.ada_max_terms  
        else:
            f_max_terms = rqaoa.max_terms 

        # If above cutoff, loop quantumly, else classically
        while n_qubits > n_cutoff:

            # Run QAOA
            q.optimize()

            # Obtain statistical results
            exp_vals_z, corr_matrix = self._exp_val_hamiltonian_termwise(q)
            # Retrieve highest expectation values according to adaptive method or schedule in custom method
            max_terms_and_stats = f_max_terms(exp_vals_z, corr_matrix, self._n_step(n_qubits, n_cutoff, counter))
            # Generate spin map
            spin_map = rqaoa.spin_mapping(problem, max_terms_and_stats)
            # Eliminate spins and redefine problem
            new_problem, spin_map = rqaoa.redefine_problem(problem, spin_map)

            # Extract final set of eliminations with correct dependencies and update tracker
            eliminations = {(spin_map[spin][1],spin):spin_map[spin][0] for spin in sorted(spin_map.keys()) if spin != spin_map[spin][1]}
            elimination_tracker.append(eliminations)

            # Extract new number of qubits
            n_qubits = new_problem.n

            # Save qaoa object and new problem
            qaoa_steps.append(copy.deepcopy(q))
            problem_steps.append(copy.deepcopy(new_problem))

            # problem is updated
            problem = new_problem
            
            # Compile qaoa with the problem
            q.compile(problem, verbose=False)

            # Add one step to the counter
            counter += 1

        # Solve the new problem classically
        cl_energy, cl_ground_states = ground_state_hamiltonian(problem.hamiltonian)

        # Retrieve full solutions including eliminated spins and their energies
        full_solutions = rqaoa.final_solution(
            elimination_tracker, cl_ground_states, self.problem.hamiltonian)

        # Compute description dictionary containing all the information            
        self.results['solution'] = full_solutions
        self.results['classical_output'] = {'minimum_energy': cl_energy,  'optimal_states': cl_ground_states}
        self.results['elimination_rules'] = elimination_tracker
        self.results['schedule'] = [len(max_tc) for max_tc in elimination_tracker]
        self.results['intermediate_steps'] = [{'QUBO': problem, 'QAOA': qaoa} for qaoa, problem in zip(qaoa_steps, problem_steps)]
        self.results['number_steps'] = counter - self.rqaoa_parameters.counter 

        if verbose:
            print(f'RQAOA optimization completed.')

        return 


    def _exp_val_hamiltonian_termwise(self, q):
        """
        Private method to call the exp_val_hamiltonian_termwise function taking the data from
        the QAOA object _q. 
        It eturns what the exp_val_hamiltonian_termwise function returns.
        """

        variational_params = q.variate_params
        qaoa_backend = q.backend
        cost_hamiltonian = q.cost_hamil
        mixer_type = q.circuit_properties.mixer_hamiltonian
        p = q.circuit_properties.p
        qaoa_optimized_angles = q.results.optimized['optimized angles']
        qaoa_optimized_counts = q.results.get_counts(q.results.optimized['optimized measurement outcomes'])
        analytical = isinstance(qaoa_backend, QAOABaseBackendStatevector)
    
        return exp_val_hamiltonian_termwise(variational_params, 
                qaoa_backend, cost_hamiltonian, mixer_type, p, qaoa_optimized_angles, 
                qaoa_optimized_counts, analytical=analytical)


    def _n_step(self, n_qubits, n_cutoff, counter):
        """
        Private method that returns the n_max value in case of adaptive or the number of eliminations according 
        to the schedule and the counter in case of custom method.
        """

        if self.rqaoa_parameters.rqaoa_type == "adaptive":
            # Number of spins to eliminate according the schedule
            n = self.rqaoa_parameters.n_max
        else:
            # max Number of spins to eliminate
            n = self.rqaoa_parameters.steps[counter]

        # If the step eliminates more spins than available, reduce step to match cutoff
        return (n_qubits - n_cutoff) if (n_qubits - n_cutoff) < n else n


#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

class RQAOAResults(dict):
    """
    A class to handle the results of RQAOA workflows
    It stores the results of the RQAOA optimization as a dictionary. With some custom methods.

    TODO :  make it similar to the QAOA results class, some plotting functions could be added too
            QAQAOAResults class is in optimizers/result.py
    """

    def get_solution(self):
        """
        Returns the solution of the optimization.
        """
        return self['solution']

    def get_step(self, i):
        """
        Returns the QUBO problem and QAOA of the i-th intermidate step of the optimization.
        """
        return self['intermediate_steps'][i]

    def get_qaoa_step(self, i):
        """
        Returns the i-th qaoa step of the RQAOA.
        """
        return self['intermediate_steps'][i]['QAOA']

    def get_qaoa_step_optimized_angles(self, i):
        """
        Returns the optimized angles of the i-th qaoa step of the RQAOA.
        """
        return self.get_qaoa_step(i).results.optimized['optimized angles']

    def get_problem_step(self, i):
        """
        Returns the QUBO problem in the i-th step of the RQAOA.
        """
        return self['intermediate_steps'][i]['QUBO']

    def get_hamiltonian_step(self, i):
        """
        Returns the Hamiltonian of the i-th step of the RQAOA.
        """
        return self.get_problem_step(i).hamiltonian


#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# General Imports
from ...basebackend import QAOABaseBackendParametric, QAOABaseBackendShotBased, QAOABaseBackendStatevector
from ...qaoa_parameters.baseparams import QAOACircuitParams, QAOAVariationalBaseParams
from ...utilities import flip_counts
from ...cost_function import cost_function
from ...qaoa_parameters.gatemap import (RXGateMap, RYGateMap, RZGateMap, RXXGateMap,
                                          RYYGateMap, RZZGateMap, RZXGateMap)

import numpy as np
from typing import Union, List, Tuple, Optional

# IBM Qiskit imports
from qiskit import QuantumCircuit, QuantumRegister
from qiskit.providers.aer import AerSimulator
from qiskit.providers.aer.noise import NoiseModel
from qiskit.opflow.primitive_ops import PauliSumOp
from qiskit.quantum_info import Statevector
from qiskit.circuit import Parameter

"""
QASM Simluator can be used for different simulation purposes with different Simulation methods
    - supports different error models
    - supports incluing real IBM backend error models
"""

class QAOAQiskitBackendShotBasedSimulator(QAOABaseBackendShotBased, QAOABaseBackendParametric):
    """
    Local Shot-based simulators offered by Qiskit

    Parameters
    ----------
    circuit_params: `QAOACircuitParams`
        An object of the class ``QAOACircuitParams`` which contains information on 
        circuit construction and depth of the circuit.
    
    n_shots: `int`
        The number of shots to be taken for each circuit.
    
    seed_simulator: `int`
        Pseudorandom list of numbers of a seed

    prepend_state: `QuantumCircuit`
        The state prepended to the circuit.

    append_state: `QuantumCircuit`
        The state appended to the circuit.

    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the 
        QAOA part of the circuit.

    cvar_alpha: `float`
        The value of alpha for the CVaR cost function.
    
    qiskit_simulation_method: `str`
        The method to be used for the simulation.

    noise_model: `NoiseModel`
        The Qiskit noise model to be used for the simulation.
    """    
    QISKIT_GATEMAP_LIBRARY = [RXGateMap, RYGateMap, RZGateMap, RXXGateMap,
                                RYYGateMap, RZZGateMap, RZXGateMap]


    def __init__(self,
                 circuit_params: QAOACircuitParams,
                 n_shots: int,
                 prepend_state: Optional[QuantumCircuit],
                 append_state: Optional[QuantumCircuit],
                 init_hadamard: bool,
                 cvar_alpha: float,
                 qiskit_simulation_method: str = 'automatic',
                 seed_simulator: Optional[int] = None,
                 noise_model: Optional[NoiseModel] = None):
        
        QAOABaseBackendShotBased.__init__(self,circuit_params,
                                          n_shots,
                                          prepend_state,
                                          append_state,
                                          init_hadamard,
                                          cvar_alpha)

        self.qureg = QuantumRegister(self.n_qubits)
        self.qubit_layout = self.circuit_params.qureg
        
        if self.prepend_state:
            assert self.n_qubits >= len(prepend_state.qubits), "Cannot attach a bigger circuit " \
                                                                "to the QAOA routine"
        # options = {"seed_simulator":1}
        self.backend_simulator = AerSimulator(method=qiskit_simulation_method.lower(),
                                              noise_model = noise_model, seed_simulator=seed_simulator)
        # For parametric circuits
        self.parametric_circuit = self.parametric_qaoa_circuit

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> QuantumCircuit:
        """
        The final QAOA circuit to be executed on the QPU.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
        
        Returns
        -------
        qaoa_circuit: `QuantumCircuit`
            The final QAOA circuit after binding angles from variational parameters.
        """
        angles_list = self.obtain_angles_for_pauli_list(self.abstract_circuit, params)
        memory_map = dict(zip(self.qiskit_parameter_list, angles_list))
        new_parametric_circuit = self.parametric_circuit.bind_parameters(memory_map)
        return new_parametric_circuit
    
    @property
    def parametric_qaoa_circuit(self) -> QuantumCircuit:
        """
        Creates a parametric QAOA circuit, given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit.
        """
        # self.reset_circuit()
        parametric_circuit = QuantumCircuit(self.qureg)

        if self.prepend_state:
            parametric_circuit = parametric_circuit.compose(self.prepend_state)
        # Initial state is all |+>
        if self.init_hadamard:
            parametric_circuit.h(self.qureg)
        
        self.qiskit_parameter_list=[]
        for each_gate in self.abstract_circuit:
            angle_param = Parameter(str(each_gate.pauli_label))
            self.qiskit_parameter_list.append(angle_param)
            each_gate.rotation_angle = angle_param
            if type(each_gate) in self.QISKIT_GATEMAP_LIBRARY:
                decomposition = each_gate.decomposition('trivial')
            else: 
                decomposition = each_gate.decomposition('standard')
            # Create Circuit
            for each_tuple in decomposition:
                low_gate = each_tuple[0]()
                parametric_circuit = low_gate.apply_ibm_gate(*each_tuple[1],parametric_circuit)
        
        if self.append_state:
            parametric_circuit = parametric_circuit.compose(self.append_state)
        parametric_circuit.measure_all()

        return parametric_circuit

    def get_counts(self, params: QAOAVariationalBaseParams) -> dict:
        """
        Returns the counts of the final QAOA circuit after binding angles from variational parameters.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
        
        Returns
        -------
        counts: `dict`
            The counts of the final QAOA circuit after binding angles from variational parameters.
        """
        qaoa_circuit = self.qaoa_circuit(params)
        counts = self.backend_simulator.run(qaoa_circuit, shots=self.n_shots).result().get_counts()
        flipped_counts = flip_counts(counts)
        self.measurement_outcomes = flipped_counts
        return flipped_counts

    def circuit_to_qasm(self):
        """
        A method to convert the QAOA circuit to QASM.
        """
        raise NotImplementedError()
#         qasm_circuit = self.parametric_circuit.qasm()
#         return qasm_circuit

    def reset_circuit(self):
        raise NotImplementedError()
    

class QAOAQiskitBackendStatevecSimulator(QAOABaseBackendStatevector, QAOABaseBackendParametric):
    """
    Local Shot-based simulators offered by Qiskit

    Parameters
    ----------
    circuit_params: `QAOACircuitParams`
        An object of the class ``QAOACircuitParams`` which contains information on 
        circuit construction and depth of the circuit.
    
    n_shots: `int`
        The number of shots to be taken for each circuit.

    prepend_state: `np.ndarray` or `QuantumCircuit`
        The state prepended to the circuit.

    append_state: `QuantumCircuit or np.ndarray`
        The state appended to the circuit.

    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the 
        QAOA part of the circuit.

    cvar_alpha: `float`
        The value of alpha for the CVaR cost function.
    """
    QISKIT_GATEMAP_LIBRARY = [RXGateMap, RYGateMap, RZGateMap, RXXGateMap,
                                RYYGateMap, RZZGateMap, RZXGateMap]

    def __init__(self,
                 circuit_params: QAOACircuitParams,
                 prepend_state: Optional[Union[np.ndarray,QuantumCircuit]],
                 append_state: Optional[Union[np.ndarray,QuantumCircuit]],
                 init_hadamard: bool,
                 cvar_alpha: float = 1):
        
        QAOABaseBackendStatevector.__init__(self, circuit_params,
                                            prepend_state,
                                            append_state,
                                            init_hadamard,
                                            cvar_alpha)

        assert cvar_alpha == 1,  "Please use the shot-based simulator for simulations with cvar_alpha < 1"

        self.qureg = QuantumRegister(self.n_qubits)
        self.qubit_layout = self.circuit_params.qureg
        
        if self.prepend_state:
            assert self.n_qubits >= len(prepend_state.qubits), "Cannot attach a bigger circuit " \
                                                                "to the QAOA routine"

        # For parametric circuits
        self.parametric_circuit = self.parametric_qaoa_circuit
        self.qiskit_cost_hamil = self.qiskit_cost_hamiltonian
        self.qiskit_cost_hamil_sq = self.qiskit_cost_hamil**2

    @property
    def qiskit_cost_hamiltonian(self):
        """
        The qiskit cost hamiltonian for the QAOA circuit represented
        as a `PauliSumOp` object.
        """
        cost_hamil = self.cost_hamiltonian
        n_qubits = cost_hamil.n_qubits
        pauli_strings_list = ['I'*n_qubits]*len(cost_hamil.terms)
        for i,pauli_op in enumerate(cost_hamil.terms):
            pauli_term = list(pauli_strings_list[i])
            for pauli,qubit in zip(pauli_op.pauli_str,pauli_op.qubit_indices):
                pauli_term[qubit] = pauli
            #reverse pauli_strings because qiskit supports little endian notation
            pauli_strings_list[i] = ''.join(str(term) for term in pauli_term)[::-1]
                
        pauli_strings_list.append('I'*n_qubits)
        pauli_coeffs = cost_hamil.coeffs
            
        qiskit_pauli_op = [[pauli_strings,coeff] for pauli_strings,coeff in zip(pauli_strings_list,pauli_coeffs)]
        qiskit_pauli_op.append(['I'*n_qubits,cost_hamil.constant])
        qiskit_cost_hamil = PauliSumOp.from_list(qiskit_pauli_op)
        return qiskit_cost_hamil

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> QuantumCircuit:
        """
        The final QAOA circuit to be executed on the QPU.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
        
        Returns
        -------
        qaoa_circuit: `QuantumCircuit`
            The final QAOA circuit after binding angles from variational parameters.
        """
        angles_list = self.obtain_angles_for_pauli_list(self.abstract_circuit, params)
        memory_map = dict(zip(self.qiskit_parameter_list, angles_list))
        new_parametric_circuit = self.parametric_circuit.bind_parameters(memory_map)
        return new_parametric_circuit
    
    @property
    def parametric_qaoa_circuit(self) -> QuantumCircuit:
        """
        Creates a parametric QAOA circuit, given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit. To do this, you will need to subsequently execute the command self.eng.flush().

        Parameters
        ----------
            params:
                Object of type QAOAVariationalBaseParams
        """
        # self.reset_circuit()
        parametric_circuit = QuantumCircuit(self.qureg)

        if self.prepend_state:
            parametric_circuit = parametric_circuit.compose(self.prepend_state)
        # Initial state is all |+>
        if self.init_hadamard:
            parametric_circuit.h(self.qureg)
        
        self.qiskit_parameter_list=[]
        for each_gate in self.abstract_circuit:
            angle_param = Parameter(str(each_gate.pauli_label))
            self.qiskit_parameter_list.append(angle_param)
            each_gate.rotation_angle = angle_param
            if type(each_gate) in self.QISKIT_GATEMAP_LIBRARY:
                decomposition = each_gate.decomposition('trivial')
            else: 
                decomposition = each_gate.decomposition('standard')
            # Create Circuit
            for each_tuple in decomposition:
                low_gate = each_tuple[0]()
                parametric_circuit = low_gate.apply_ibm_gate(*each_tuple[1],parametric_circuit)

        if self.append_state:
            parametric_circuit = parametric_circuit.compose(self.append_state)

        return parametric_circuit

    def wavefunction(self,
                     params: QAOAVariationalBaseParams) -> Union[List[complex],np.ndarray]:
        """
        Get the wavefunction of the state produced by the parametric circuit.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        wf: `List[complex]` or `np.ndarray[complex]`
            A list of the wavefunction amplitudes.
         """
        ckt = self.qaoa_circuit(params)
        wf = Statevector(ckt).data
        self.measurement_outcomes = wf
        return wf

    def expectation(self,
                    params: QAOAVariationalBaseParams) -> float:
        """
        Compute the expectation value w.r.t the Cost Hamiltonian

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes, containing 
            variable parameters.

        Returns
        -------
        `float`
            expectation value of cost operator wrt to quantum state produced by QAOA circuit
        """
        ckt = self.qaoa_circuit(params)
        output_wf = Statevector(ckt)
        self.measurement_outcomes = output_wf.data
        cost = np.real(output_wf.expectation_value(self.qiskit_cost_hamil))
        return cost

    def expectation_w_uncertainty(self, 
                                  params: QAOAVariationalBaseParams) -> Tuple[float, float]:
        """
        Compute the expectation value w.r.t the Cost Hamiltonian and its uncertainty

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes, containing 
            variable parameters.

        Returns
        -------
        `Tuple[float]`
            expectation value and its uncertainty of cost operator wrt 
            to quantum state produced by QAOA circuit.
        """
        ckt = self.qaoa_circuit(params)
        output_wf = Statevector(ckt)
        self.measurement_outcomes = output_wf.data
        cost = np.real(Statevector(ckt).expectation_value(self.qiskit_cost_hamil))
        cost_sq = np.real(Statevector(ckt).expectation_value(self.qiskit_cost_hamil_sq))
        
        uncertainty = np.sqrt(cost_sq - cost**2)
        return (cost, uncertainty)

    def reset_circuit(self):
        """
        Reset self.circuit after performing a computation

        TODO: Check if only the first instruction is needed. Two might
              be redundant
        """
        raise NotImplementedError()

    def circuit_to_qasm(self):
        """
        """
        raise NotImplementedError()



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Tuple
import numpy as np
from pyquil import Program, gates
from pyquil.api import WavefunctionSimulator

from ...basebackend import QAOABaseBackendStatevector
from ...qaoa_parameters.baseparams import QAOACircuitParams, QAOAVariationalBaseParams
from ...qaoa_parameters.gatemap import (
    RXGateMap, RYGateMap, RZGateMap)
from ...cost_function import cost_function


class QAOAPyQuilWavefunctionSimulatorBackend(QAOABaseBackendStatevector):
    """
    A local Wavefunction simulator backend for the PyQuil service provider
    """
    PYQUIL_PAULIGATE_LIBRARY = [RXGateMap, RYGateMap, RZGateMap]

    def __init__(self,
                 circuit_params: QAOACircuitParams,
                 prepend_state: Program = None,
                 append_state: Program = None,
                 init_hadamard: bool = True,
                 cvar_alpha: float = 1):

        QAOABaseBackendStatevector.__init__(self, circuit_params, prepend_state,
                                            append_state, init_hadamard,
                                            cvar_alpha)

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> Program:
        """
        Creates a QAOA circuit (pyquil.Program object), given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit. To do this, you will need to subsequently execute the command self.eng.flush().

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        `pyquil.Program`
                A pyquil.Program object.
        """
        self.assign_angles(params)

        circuit = Program()
        if self.prepend_state:
            circuit += self.prepend_state

        # Initial state is all |+>
        if self.init_hadamard:
            for i in range(self.n_qubits):
                circuit += gates.H(i)

        # create a list of gates in order of application on quantum circuit
        low_level_gate_list = []
        for i,each_gate in enumerate(self.abstract_circuit):
            if type(each_gate) in self.PYQUIL_PAULIGATE_LIBRARY:
                decomposition = each_gate.decomposition('trivial')
            else:
                decomposition = each_gate.decomposition('standard')
            # using the list above, construct the circuit
            for each_tuple in decomposition:
                low_gate = each_tuple[0]()
                circuit = low_gate.apply_pyquil_gate(*each_tuple[1],circuit)

        if self.append_state:
            circuit += self.append_state

        return circuit
    
    def wavefunction(self,
                     params: QAOAVariationalBaseParams):
        """
        Get the wavefunction of the state produced by the parametric circuit.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
                The QAOA parameters - an object of one of the parameter classes, containing 
                the variational parameters (angles).

        Returns
        -------
        wf: `List[complex]`
                pyquil Wavefunction object.
        """
        program = self.qaoa_circuit(params)

        wf_sim = WavefunctionSimulator()
        wf = wf_sim.wavefunction(program)
        self.measurement_outcomes = wf.amplitudes
        return wf

    def expectation(self, params: QAOAVariationalBaseParams) -> float:
        """
        Compute the expectation value w.r.t the Cost Hamiltonian

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
                The QAOA parameters - an object of one of the parameter classes, containing 
                variable parameters.

        Returns
        -------
        `float`
                expectation value of cost operator wrt to quantum state produced by QAOA circuit
        """
        prob_dict = self.probability_dict(params)
        cost = cost_function(
            prob_dict, self.circuit_params.cost_hamiltonian, self.cvar_alpha)
        return cost

    def expectation_w_uncertainty(self,
                                  params: QAOAVariationalBaseParams) -> Tuple[float, float]:
        """
        Compute the expectation value w.r.t the Cost Hamiltonian and its uncertainty

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
                The QAOA parameters - an object of one of the parameter classes, containing 
                variable parameters.

        Returns
        -------
        `Tuple[float]`
                expectation value and its uncertainty of cost operator wrt 
                to quantum state produced by QAOA circuit.
        """
        prob_dict = self.probability_dict(params)
        cost = cost_function(
            prob_dict, self.circuit_params.cost_hamiltonian, self.cvar_alpha)
        cost_sq = cost_function(prob_dict,
                                self.circuit_params.cost_hamiltonian.hamiltonian_squared,
                                self.cvar_alpha)

        uncertainty = np.sqrt(cost_sq - cost**2)

        return (cost, uncertainty)

    def circuit_to_qasm(self):
        raise NotImplementedError()

    def reset_circuit(self):
        raise NotImplementedError()



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains backend specific codes for implementing QAOA on supported simulators
and QPU backends

Currently Supported:
	Qiskit:
		QASM Simulator
		Statevector Simulator
		IBMQ available QPUs
	PyQuil:
		Rigetti QPUs
		Statevector Simulator
	vectorized:
		Fast numpy native Statevector Simulator
"""
from .qpus.qaoa_qiskit_qpu import *
from .qpus.qaoa_pyquil_qpu import *
from .qpus.qaoa_braket_qpu import *

from .simulators.qaoa_qiskit_sim import *
from .simulators.qaoa_pyquil_sim import *
from .simulators.qaoa_vectorized import *



from __future__ import annotations

import numpy as np

from copy import deepcopy
from .basebackend import QAOABaseBackendShotBased
from .optimizers.logger_vqa import Logger
from .basebackend import QAOABaseBackend
from .qaoa_parameters.baseparams import QAOAVariationalBaseParams


def log_qfim_evals(logger: Logger) -> Logger:
    
    current_total_eval = logger.func_evals.best[0]
    current_total_eval += 1
    current_qfim_eval = logger.qfim_func_evals.best[0]
    current_qfim_eval += 1
    logger.log_variables({'func_evals': current_total_eval, 
                          'qfim_func_evals': current_qfim_eval})
    
    return logger

def qfim(backend_obj: QAOABaseBackend, 
         params: QAOAVariationalBaseParams, 
         logger: Logger, 
         eta: float = 0.00000001):
    
        """
        Returns a callable qfim_fun(args) that computes the quantum fisher information matrix at `args` according to :
        $$[QFI]_{ij} = Re(<∂iφ|∂jφ>) − <∂iφ|φ><φ|∂jφ>$$.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters as a 1D array (derived from an object of one of
            the parameter classes, containing hyperparameters and variable parameters).

        eta: `float`
            The infinitesimal shift used to compute `|∂jφ>`, the partial derivative 
            of the wavefunction w.r.t a parameter. 

        Returns
        -------
        qfim_array:
            The quantum fisher information matrix, a 2p*2p symmetric square matrix 
            with elements [QFI]_ij = Re(<∂iφ|∂jφ>) − <∂iφ|φ><φ|∂jφ>.
        """
        
        if isinstance(backend_obj, QAOABaseBackendShotBased):
            raise NotImplementedError("QFIM computation is not currently available on shot-based")
        
        psi = backend_obj.wavefunction(params)
        log_qfim_evals(logger)
        
        qfim_array = np.zeros((len(params), len(params)))

        copied_params = deepcopy(params)

        def qfim_fun(args):

            for i in range(len(args)):
                for j in range(i+1):
                    vi, vj = np.zeros(len(args)), np.zeros(len(args))
                    vi[i] = eta
                    vj[j] = eta

                    copied_params.update_from_raw(args + vi)
                    wavefunction_plus_i = np.array(
                        backend_obj.wavefunction(copied_params))
                    log_qfim_evals(logger)

                    copied_params.update_from_raw(args - vi)
                    wavefunction_minus_i = np.array(
                        backend_obj.wavefunction(copied_params))
                    log_qfim_evals(logger)

                    copied_params.update_from_raw(args + vj)
                    wavefunction_plus_j = np.array(
                        backend_obj.wavefunction(copied_params))
                    log_qfim_evals(logger)

                    copied_params.update_from_raw(args - vj)
                    wavefunction_minus_j = np.array(
                        backend_obj.wavefunction(copied_params))
                    log_qfim_evals(logger)

                    di_psi = (wavefunction_plus_i - wavefunction_minus_i)/eta
                    dj_psi = (wavefunction_plus_j - wavefunction_minus_j)/eta

                    qfim_array[i][j] = np.real(
                        np.vdot(di_psi, dj_psi)) - np.vdot(di_psi, psi)*np.vdot(psi, dj_psi)

                    if i != j:
                        qfim_array[j][i] = qfim_array[i][j]

            return qfim_array

        return qfim_fun


#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import abc
import numpy as np
from typing import Optional
from qiskit import IBMQ
from qiskit.providers.ibmq import IBMQAccountError
from qiskit.providers.ibmq.api.exceptions import RequestsApiError

from qcs_api_client.client import QCSClientConfiguration
from pyquil.api._engagement_manager import EngagementManager
from pyquil import get_qc

from boto3.session import Session
from botocore.exceptions import NoRegionError
from braket.aws import AwsDevice
from braket.aws.aws_session import AwsSession

SUPPORTED_LOCAL_SIMULATORS = [
    'qiskit.qasm_simulator', 'qiskit.shot_simulator',
    'qiskit.statevector_simulator','vectorized',
    'pyquil.statevector_simulator'
]


class DeviceBase(metaclass=abc.ABCMeta):
    """An object that contains the relevant information required to access 
    certain backends. Other Access Objects have to inherit from this object.
    """

    @abc.abstractmethod
    def check_connection(self) -> bool:
        """This method should allow a user to easily check if the credentials
        provided to access the remote QPU is valid.

        Returns
        -------
        bool 
            True if a connection can be established. If False, the error 
            should be logged and printable. (Not creating an exception 
            here is good for extendibility. i.e. able to try multiple
            providers without exiting the program.)
        """
        pass

class DeviceLocal(DeviceBase):
    """
    This class is a placeholder for all locally accessible devices.
    """
    def __init__(self, device_name: str):
        self.device_name = device_name
        self.device_location = 'local'

    def check_connection(self) -> bool:
        if self.device_name in SUPPORTED_LOCAL_SIMULATORS:
            return True
        else:
            return False

class DeviceQiskit(DeviceBase):
    """
    Contains the required information and methods needed to access remote
    qiskit QPUs.

    Parameters
    ----------
	available_qpus: `list`
		When connection to a provider is established, this attribute contains a list
		of backend names which can be used to access the selected backend by reinitialising
		the Access Object with the name of the available backend as input to the
		device_name parameter.
	"""

    def __init__(self, device_name: str, api_token: str,
				 hub: str, group: str, project: str):
        """A majority of the input parameters required for this can be found in
        the user's IBMQ Experience account.

        Parameters
        ----------
		device_name: `str`
			The name of the IBMQ device to be used
        api_token: `str`
            Valid IBMQ Experience Token.
        hub: `str`
            Valid IBMQ hub name.
        group: `str`
            Valid IBMQ group name. 
        project: `str`
            The name of the project for which the experimental data will be 
            saved in on IBMQ's end.
        """

        self.api_token = api_token
        self.hub = hub
        self.group = group
        self.project = project
        self.device_name = device_name
        self.device_location = 'ibmq'

        self.provider_connected = None
        self.qpu_connected = None

    def check_connection(self) -> bool:
        """
        This method should allow a user to easily check if the credentials
        provided to access the remote QPU is valid.

        If no backend was specified in initialisation of object, just runs
        a test connection without a specific backend.
        If backend was specified, checks if connection to that backend
        can be established.

        Returns
        -------
        bool
			True if successfully connected to IBMQ or IBMQ and the QPU backend
			if it was specified. False if unable to connect to IBMQ or failure
			in the attempt to connect to the specified backend.
        """

        self.provider_connected = self._check_provider_connection()

        if self.provider_connected == False:
            return self.provider_connected

        self.available_qpus = [backend.name()
                               for backend in self.provider.backends()]

        if self.device_name == '':
            return self.provider_connected

        self.qpu_connected = self._check_backend_connection()

        if self.provider_connected and self.qpu_connected:
            return True
        else:
            return False

    def _check_backend_connection(self) -> bool:
        """Private method for checking connection with backend(s).
        """

        if self.device_name in self.available_qpus:
            self.backend_device = self.provider.get_backend(self.device_name)
            return True
        else:
            print(
                f"Please choose from {self.available_qpus} for this provider")
            return False

    def _check_provider_connection(self) -> bool:
        """
        Private method for checking connection with provider.
        """

        try:
            if IBMQ.active_account() is None:
                self.provider = IBMQ.enable_account(self.api_token, hub=self.hub,
                                                    group=self.group,
                                                    project=self.project)
            elif IBMQ.active_account()['token'] != self.api_token:
                IBMQ.disable_account()
                self.provider = IBMQ.enable_account(self.api_token, hub=self.hub,
                                                    group=self.group,
                                                    project=self.project)
            else:
                self.provider = IBMQ.get_provider(hub=self.hub, group=self.group,
                                                  project=self.project)

            return True
        
        except RequestsApiError as e:
            print('The api key used was invalid: {}'.format(e))
            return False
        
        except Exception as e:
            print('An Exception has occured when trying to connect with the \
            provider: {}'.format(e))
            return False


class DevicePyquil(DeviceBase):
    """
    Contains the required information and methods needed to access remote
    Rigetti QPUs via Pyquil.
    """

    def __init__(self, device_name: str, as_qvm: bool = None, noisy: bool = None,
                 compiler_timeout: float = 20.0,
                 execution_timeout: float = 20.0,
                 client_configuration: QCSClientConfiguration = None,
                 endpoint_id: str = None,
                 engagement_manager: EngagementManager = None):
        """
        Parameters
        ----------
        device_name: str 
            The name of the desired quantum computer. This should correspond to 
            a name returned by :py:func:`list_quantum_computers`. Names ending 
            in "-qvm" will return a QVM. Names ending in "-pyqvm" will return a 
            :py:class:`PyQVM`. Names ending in "-noisy-qvm" will return a QVM 
            with a noise model. Otherwise, we will return a QPU with the given 
            name.
        as_qvm: bool 
            An optional flag to force construction of a QVM (instead of a QPU). 
            If specified and set to ``True``, a QVM-backed quantum computer will 
            be returned regardless of the name's suffix.
        noisy: bool
            An optional flag to force inclusion of a noise model. If specified 
            and set to ``True``, a quantum computer with a noise model will be 
            returned regardless of the name's suffix. The generic QVM noise 
            model is simple T1 and T2 noise plus readout error. See 
            :py:func:`~pyquil.noise.decoherence_noise_with_asymmetric_ro`. Note, 
            we currently do not support noise models based on QCS hardware; a 
            value of `True`` will result in an error if the requested QPU is a 
            QCS hardware QPU.
        compiler_timeout: float
            Time limit for compilation requests, in seconds.
        execution_timeout: float
            Time limit for execution requests, in seconds.
        client_configuration: QCSClientConfiguration
            Optional client configuration. If none is provided, a default one 
            will be loaded.
        endpoint_id: str
            Optional quantum processor endpoint ID, as used in the 
            `QCS API Docs`_.
        engagement_manager: EngagementManager
            Optional engagement manager. If none is provided, a default one will 
            be created.
        """

        self.device_name = device_name
        self.device_location = 'qcs'
        self.as_qvm = as_qvm
        self.noisy = noisy
        self.compiler_timeout = compiler_timeout
        self.execution_timeout = execution_timeout
        self.client_configuration = client_configuration
        self.endpoint_id = endpoint_id
        self.engagement_manager = engagement_manager

        self.quantum_computer = get_qc(name=self.device_name, as_qvm=self.as_qvm, noisy=self.noisy,
                                       compiler_timeout=self.compiler_timeout, execution_timeout=self.execution_timeout,
                                       client_configuration=self.client_configuration, endpoint_id=self.endpoint_id, engagement_manager=self.engagement_manager)

    def check_connection(self) -> bool:
        """This method should allow a user to easily check if the credentials
        provided to access the remote QPU is valid.

        If no device was specified in initialisation of object, just runs
        a test connection without a specific device.
        If device was specified, checks if connection to that device
        can be established.

        TODO : 
        Accessing Rigetti's QCS is currently unsupported, so this part is empty until that is figured out.
        """

        return True
    
    
class DeviceAWS(DeviceBase):
    
    """
    Contains the required information and methods needed to access QPUs hosted
    on AWS Braket.
    
    Attributes:
	available_qpus: `list`
		When connection to AWS is established, this attribute contains a list
		of device names which can be used to access the selected device by
		reinitialising the Access Object with the name of the available device
		as input to the device_name parameter.
    """
    
    def __init__(self, device_name: str, s3_bucket_name: str = None, 
                 aws_region: str = None, 
                 folder_name: str = 'openqaoa'):
        
        """Input the device arn and the name of the folder in which all the
        results for the QPU runs would be saved on the pre-defined s3 bucket. 
        Note that the user is required to authenticate through the AWS CLI 
        before being able to use this Device object.
        
        See: https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html for further details.

        Parameters
        ----------
		device_name: `str`
			The ARN string of the braket QPU/simulator to be used
        s3_bucket_name: `str`
            The name of S3 Bucket where the Braket run results will be saved.
        aws_region: `str`
            The aws region in which the QPU/simulator is located. Defaults to the
            region set in aws cli config.
        folder_name: `str`
            The name of the folder in the s3 bucket that will contain the results
            from the tasks performed in this run.
        """
        
        self.device_name = device_name
        self.device_location = 'aws'
        self.s3_bucket_name = s3_bucket_name
        self.aws_region = aws_region
        self.folder_name = folder_name
        
        self.provider_connected = None
        self.qpu_connected = None
    
    def check_connection(self) -> bool:
        
        self.provider_connected = self._check_provider_connection()

        if self.provider_connected == False:
            return self.provider_connected
        
        # Only QPUs that are available for the specified aws region on Braket 
        # will be shown. We filter out QPUs that do not work with the circuit model
        sess_devices = self.aws_session.search_devices()
        
        device_filter = np.multiply(
            [each_dict['deviceStatus'] == 'ONLINE' for each_dict in sess_devices],
            [each_dict['providerName'] != 'D-Wave Systems' for each_dict in sess_devices]
        )
        active_devices = np.array(sess_devices)[device_filter].tolist()
        
        self.available_qpus = [backend_dict['deviceArn']
                               for backend_dict in active_devices]

        if self.device_name == '':
            return self.provider_connected

        self.qpu_connected = self._check_backend_connection()

        if self.provider_connected and self.qpu_connected:
            return True
        else:
            return False
        
    def _check_backend_connection(self) -> bool:
        
        if self.device_name in self.available_qpus:
            self.backend_device = AwsDevice(self.device_name, self.aws_session)
            return True
        else:
            print(
                f"Please choose from {self.available_qpus} for this provider")
            return False
    
    def _check_provider_connection(self) -> bool:
        
        try:
            sess = Session(region_name = self.aws_region)
            self.aws_session = AwsSession(sess, default_bucket=self.s3_bucket_name)
            self.aws_region = self.aws_session.region
            self.s3_bucket_name = self.aws_session.default_bucket()
            return True
        except NoRegionError:
            self.aws_session = None
            return True
        except Exception as e:
            print('An Exception has occured when trying to connect with the provider. You are required to authenticate through the AWS CLI in order to connect to the Braket QPUs. Please check if you have properly set it up. See: https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html for further details. : {}'.format(e))
            return False


def device_class_arg_mapper(device_class:DeviceBase,
                            api_token: str = None,
                            hub: str = None,
                            group: str = None,
                            project: str = None,
                            as_qvm: bool = None,
                            noisy: bool = None,
                            compiler_timeout: float = None,
                            execution_timeout: float = None,
                            client_configuration: QCSClientConfiguration = None,
                            endpoint_id: str = None,
                            engagement_manager: EngagementManager = None,
                            device_name: str = None,
                            folder_name: str = None, 
                            s3_bucket_name:str = None, 
                            aws_region: str = None) -> dict:
    DEVICE_ARGS_MAPPER = {
        DeviceQiskit: {'api_token': api_token,
                        'hub': hub,
                        'group': group,
                        'project': project},

        DevicePyquil: {'as_qvm': as_qvm,
                        'noisy': noisy,
                        'compiler_timeout': compiler_timeout,
                        'execution_timeout': execution_timeout,
                        'client_configuration': client_configuration,
                        'endpoint_id': endpoint_id,
                        'engagement_manager': engagement_manager},
        
        DeviceAWS: {'device_name': device_name,
                    's3_bucket_name': s3_bucket_name,
                    'aws_region': aws_region,
                    'folder_name': folder_name}
    }

    final_device_kwargs = {key: value for key, value in DEVICE_ARGS_MAPPER[device_class].items()
                           if value is not None}
    return final_device_kwargs


def create_device(location: str, name: str, **kwargs):
    """
    This function returns an instance of the appropriate device class.

    Parameters
    ----------
    device_name: str
        The name of the device to be accessed.
    device_location: str
        The location of the device to be accessed.
    kwargs: dict
        A dictionary of keyword arguments to be passed to the device class.
        These will be used to initialise the device.

    Returns
    -------
    device: DeviceBase
        An instance of the appropriate device class.
    """
    location = location.lower()
    if location == 'ibmq':
        device_class = DeviceQiskit
    elif location == 'qcs':
        device_class = DevicePyquil
    elif location == 'aws':
        device_class = DeviceAWS
    elif location == 'local':
        device_class = DeviceLocal
    else:
        raise ValueError(f'Invalid device location, Choose from: {location}')

    return device_class(device_name=name, **kwargs)


#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Cost function to be used for QAOA training
from typing import Dict
from collections import OrderedDict
from .qaoa_parameters.operators import Hamiltonian
from .utilities import bitstring_energy
import numpy as np

def expectation_value_classical(counts: Dict, hamiltonian: Hamiltonian):
    """
    Evaluate the cost function, i.e. expectation value ``$$\langle|H \rangle$$``
    w.r.t to the measurements results ``counts``.

    Parameters
    ----------
    counts: dict
		The counts of the measurements.
    hamiltonian: Hamiltonian	
		The Cost Hamiltonian defined for the optimization problem
    """
    shots = sum(counts.values())
    cost = 0
    for basis_state, count in counts.items():
        cost += count*(bitstring_energy(hamiltonian, basis_state) / shots)
    return cost


def cvar_expectation_value_classical(counts: Dict, hamiltonian: Hamiltonian, alpha: float):
	"""
	CVaR computation of cost function. For the definition of the cost function, refer
	to https://arxiv.org/abs/1907.04769.

	Parameters
	----------
	counts: `dict`
		The counts of the measurements.
	hamiltonian: `Hamiltonian`
		The Cost Hamiltonian defined for the optimization problem
	alpha: `float`
		The CVaR parameter.
	"""
	assert alpha > 0 and alpha < 1, "Please specify a valid alpha value between 0 and 1"
	shots = sum(counts.values())

	cost_list = []

	#eigen-energy computation of each basis state in counts
	for basis_state, count in counts.items():
		cost_list.append(count*(bitstring_energy(hamiltonian, basis_state) / shots))

	#sort costs in ascending order
	sorted_cost_list = np.sort(np.array(cost_list))

	K = max([int(len(cost_list)*alpha),1])
	cost = sum(sorted_cost_list[:K])

	return cost


def cost_function(counts: Dict, hamiltonian: Hamiltonian, alpha: float = 1):
	"""
	The cost function to be used for QAOA training.

	Parameters
	----------
	counts: `dict`
		The counts of the measurements.
	hamiltonian: `Hamiltonian`
		The Cost Hamiltonian defined for the optimization problem
	alpha: `float`
		The CVaR parameter.
	"""
	if alpha == 1:
		return expectation_value_classical(counts, hamiltonian)
	else:
		return cvar_expectation_value_classical(counts, hamiltonian, alpha)
